# -*- coding: utf-8 -*-

"""Copyright (C) 2009-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

# See the user manual for a description of how to define rulesets.
# Names and descriptions must be english and may only contain ascii chars.
# The KDE translation teams will "automatically" translate name and
# description into many languages.

from typing import Optional

from rule import PredefinedRuleset
from mi18n import i18nE, i18n


class ClassicalChinese(PredefinedRuleset):

    """classical chinese rules, standard rules. Serves as a basis
    for local variants. This should be defined such that the
    sum of the differences to the local variants is minimized."""

    def __init__(self, name:Optional[str]=None) ->None:
        PredefinedRuleset.__init__(
            self,
            name or i18nE('Classical Chinese standard'))

    def _initRuleset(self) ->None:
        """set the description"""
        assert isinstance(self.raw_data, str)
        self.name = self.raw_data
        self.description = i18n('Classical Chinese')

    def addManualRules(self) ->None:
        """those are actually winner rules but in the kajongg scoring mode they must be selected manually"""
        # applicable only if we have a concealed meld and a declared kong:
        self.winnerRules.createRule(
            'Last Tile Taken from Dead Wall',
            'FLastTileFromDeadWall||Olastsource=e', doubles=1,
            description=i18n('The dead wall is also called kong box: The last 16 tiles of the wall '
                             'used as source of replacement tiles'))
        self.winnerRules.createRule(
            'Last Tile is Last Tile of Wall',
            'FIsLastTileFromWall||Olastsource=z', doubles=1,
            description=i18n('Winner said Mah Jong with the last tile taken from the living end of the wall'))
        self.winnerRules.createRule(
            'Last Tile is Last Tile of Wall Discarded',
            'FIsLastTileFromWallDiscarded||Olastsource=Z', doubles=1,
            description=i18n('Winner said Mah Jong by claiming the last tile taken from the living end of the '
                             'wall, discarded by another player'))
        self.winnerRules.createRule(
            'Robbing the Kong', r'FRobbingKong||Olastsource=k', doubles=1,
            description=i18n('Winner said Mah Jong by claiming the 4th tile of a kong another player '
                             'just declared'), debug=True)
        self.winnerRules.createRule(
            'Mah Jongg with Original Call',
            'FMahJonggWithOriginalCall||Oannouncements=a', doubles=1,
            description=i18n(
                'Just before the first discard, a player can declare Original Call meaning she needs only one '
                'tile to complete the hand and announces she will not alter the hand in any way (except bonus tiles)'))
        self.winnerRules.createRule(
            'Dangerous Game', 'FDangerousGame||Opayforall',
            description=i18n('In some situations discarding a tile that has a high chance to help somebody to win '
                             'is declared to be dangerous, and if that tile actually makes somebody win, the discarder '
                             'pays the winner for all'))
        self.winnerRules.createRule(
            'Twofold Fortune', 'FTwofoldFortune||Oannouncements=t',
            limits=1, description=i18n('Kong after Kong: Declare Kong and a second Kong with the replacement '
                                       'tile and Mah Jong with the second replacement tile'))
        # limit hands:
        self.winnerRules.createRule(
            'Blessing of Heaven', 'FBlessingOfHeaven||Olastsource=1', limits=1,
            description=i18n('East says Mah Jong with the unmodified dealt tiles'))
        self.winnerRules.createRule(
            'Blessing of Earth', 'FBlessingOfEarth||Olastsource=1', limits=1,
            description=i18n('South, West or North says Mah Jong with the first tile discarded by East'))
        self.winnerRules.createRule(
            'East won nine times in a row', 'FEastWonNineTimesInARow||Orotate', limits=1,
            description=i18n('If that happens, East gets a limit score and the winds rotate'))

    def addPenaltyRules(self) ->None:
        """as the name says"""
        self.penaltyRules.createRule(
            'False Naming of Discard, Claimed for Mah Jongg and False Declaration of Mah Jongg',
            'Oabsolute payers=2 payees=2', points=-300)

    def addHandRules(self) ->None:
        """as the name says"""
        self.handRules.createRule(
            'Own Flower and Own Season',
            'FOwnFlowerOwnSeason',
            doubles=1)
        self.handRules.createRule('All Flowers', 'FAllFlowers', doubles=1)
        self.handRules.createRule('All Seasons', 'FAllSeasons', doubles=1)
        self.handRules.createRule(
            'Three Concealed Pongs',
            'FThreeConcealedPongs',
            doubles=1)
        self.handRules.createRule(
            'Long Hand', r'FLongHand||Oabsolute', points=0,
            description=i18n('The hand contains too many tiles'))

    def addParameterRules(self) ->None:
        """as the name says"""
        self.parameterRules.createRule(
            'Tiles in Hand',
            'intdealtTiles',
            description='# of tiles dealt for every player',
            parameter=13)
        self.parameterRules.createRule(
            'Points Needed for Mah Jongg',
            'intminMJPoints',
            parameter=0)
        self.parameterRules.createRule(
            'Minimum number of doubles needed for Mah Jongg',
            'intminMJDoubles', parameter=0)
        self.parameterRules.createRule(
            'Points for a Limit Hand',
            'intlimit||Omin=1',
            parameter=500)
        self.parameterRules.createRule(
            'Play with the roof off', 'boolroofOff', parameter=False,
            description=i18n('Play with no upper scoring limit'))
        self.parameterRules.createRule(
            'Claim Timeout',
            'intclaimTimeout',
            parameter=10)
        self.parameterRules.createRule(
            'Size of Kong Box', 'intkongBoxSize', parameter=16,
            description=i18n('The Kong Box is used for replacement tiles when declaring kongs'))
        self.parameterRules.createRule(
            'Play with Bonus Tiles', 'boolwithBonusTiles', parameter=True,
            description=i18n('Bonus tiles increase the luck factor'))
        self.parameterRules.createRule(
            'Minimum number of rounds in game',
            'intminRounds',
            parameter=4)
        self.parameterRules.createRule(
            'number of allowed chows', 'intmaxChows', parameter=4,
            description=i18n('The number of chows a player may build'))
        self.parameterRules.createRule(
            'must declare calling hand',
            'boolmustDeclareCallingHand', parameter=False,
            description=i18n('Mah Jongg is only allowed after having declared to have a calling hand'))
        self.parameterRules.createRule(
            'Standard Rotation',
            'FStandardRotation||Orotate||Ointernal')
        self.parameterRules.createRule(
            'Discard tiles ordered',
            'booldiscardTilesOrdered', parameter=False,
            description=i18n('Discard tiles ordered from top left'))
        self.parameterRules.createRule(
            'Leave holes for claimed discards',
            'booldiscardTilesOrderedLeaveHole', parameter=True,
            description=i18n(
                'Only when discarding tiles ordered: Do not place the next discard where the claimed discard was'))

    def loadRules(self) ->None:
        """define the rules"""
        self.addParameterRules()  # must be first!
        self.addPenaltyRules()
        self.addHandRules()
        self.addManualRules()
        self.winnerRules.createRule(
            'Last Tile Completes Pair of 2..8',
            'FLastTileCompletesPairMinor',
            points=2)
        self.winnerRules.createRule(
            'Last Tile Completes Pair of Terminals or Honors',
            'FLastTileCompletesPairMajor', points=4)
        self.winnerRules.createRule(
            'Last Tile is Only Possible Tile',
            'FLastOnlyPossible',
            points=2)
        self.winnerRules.createRule(
            'Won with Last Tile Taken from Wall',
            'FLastFromWall',
            points=2)

        self.winnerRules.createRule(
            'Zero Point Hand', 'FZeroPointHand', doubles=1,
            description=i18n('The hand has 0 basis points excluding bonus tiles'))
        self.winnerRules.createRule('No Chow', 'FNoChow', doubles=1)
        self.winnerRules.createRule(
            'Only Concealed Melds',
            'FOnlyConcealedMelds',
            doubles=1)
        self.winnerRules.createRule(
            'False Color Game', 'FFalseColorGame', doubles=1,
            description=i18n('Only same-colored tiles (only bamboo/stone/character) '
                             'plus any number of winds and dragons'))
        self.winnerRules.createRule(
            'True Color Game', 'FTrueColorGame', doubles=3,
            description=i18n('Only same-colored tiles (only bamboo/stone/character)'))
        self.winnerRules.createRule(
            'Concealed True Color Game', 'FConcealedTrueColorGame',
            limits=1, description=i18n('All tiles concealed and of the same suit, no honors'))
        self.winnerRules.createRule(
            'Only Terminals and Honors', 'FOnlyMajors', doubles=1,
            description=i18n('Only winds, dragons, 1 and 9'))
        self.winnerRules.createRule('Only Honors', 'FOnlyHonors', limits=1,
                                    description=i18n('Only winds and dragons'))
        self.winnerRules.createRule(
            'Hidden Treasure', 'FHiddenTreasure', limits=1,
            description=i18n('Only hidden Pungs or Kongs, last tile from wall'))
        self.winnerRules.createRule(
            'Heads and Tails', 'FAllTerminals', limits=1,
            description=i18n('Only 1 and 9'))
        self.winnerRules.createRule(
            'Fourfold Plenty', 'FFourfoldPlenty', limits=1,
            description=i18n('4 Kongs'))
        self.winnerRules.createRule(
            'Three Great Scholars', 'FThreeGreatScholars', limits=1,
            description=i18n('3 Pungs or Kongs of dragons'))
        self.winnerRules.createRule('Four Blessings Hovering over the Door',
                                    'FFourBlessingsHoveringOverTheDoor', limits=1,
                                    description=i18n('4 Pungs or Kongs of winds'))
        self.winnerRules.createRule('Imperial Jade', 'FAllGreen', limits=1,
                                    description=i18n('Only green tiles: Green dragon and Bamboo 2,3,4,6,8'))
        self.winnerRules.createRule('Gathering the Plum Blossom from the Roof',
                                    'FGatheringPlumBlossomFromRoof', limits=1,
                                    description=i18n('Mah Jong with stone 5 from the dead wall'))
        self.winnerRules.createRule(
            'Plucking the Moon from the Bottom of the Sea', 'FPluckingMoon', limits=1,
            description=i18n('Mah Jong with the last tile from the wall being a stone 1'))
        self.winnerRules.createRule(
            'Scratching a Carrying Pole', 'FScratchingPole', limits=1,
            description=i18n('Robbing the Kong of bamboo 2'))

        # only hands matching an mjRule can win. Keep this list as short as
        # possible. If a special hand matches the standard pattern, do not put it here
        # All mjRule functions must have a winningTileCandidates() method
        self.mjRules.createRule(
            'Standard Mah Jongg',
            'FStandardMahJongg',
            points=20)
        # option internal makes it not show up in the ruleset editor
        self.mjRules.createRule(
            'Nine Gates', 'FNineGates', limits=1,
            description=i18n(
                'A concealed hand in one color 1112345678999 plus last tile of this suit (from wall or discarded)'))
        self.mjRules.createRule(
            'Thirteen Orphans', 'FThirteenOrphans||Omayrobhiddenkong', limits=1,
            description=i18n('13 single tiles: All dragons, winds, 1, 9 and a 14th tile building a pair '
                             'with one of them'))

        # doubling melds:
        self.meldRules.createRule('Pung/Kong of Dragons', 'FDragonPungKong',
                                  explainTemplate='{meldName}', doubles=1)
        self.meldRules.createRule('Pung/Kong of Own Wind', 'FOwnWindPungKong',
                                  explainTemplate='{meldType} of Own Wind ({value})', doubles=1)
        self.meldRules.createRule(
            'Pung/Kong of Round Wind', 'FRoundWindPungKong',
            explainTemplate='{meldType} of Round Wind ({value})', doubles=1)

        # exposed melds:
        self.meldRules.createRule('Exposed Kong', 'FExposedMinorKong',
                                  explainTemplate='{meldName}', points=8)
        self.meldRules.createRule(
            'Exposed Kong of Terminals', 'FExposedTerminalsKong',
            explainTemplate='{meldName}', points=16)
        self.meldRules.createRule(
            'Exposed Kong of Honors', 'FExposedHonorsKong',
            explainTemplate='{meldName}', points=16)

        self.meldRules.createRule('Exposed Pung', 'FExposedMinorPung',
                                  explainTemplate='{meldName}', points=2)
        self.meldRules.createRule(
            'Exposed Pung of Terminals', 'FExposedTerminalsPung',
            explainTemplate='{meldName}', points=4)
        self.meldRules.createRule(
            'Exposed Pung of Honors', 'FExposedHonorsPung',
            explainTemplate='{meldName}', points=4)

        # concealed melds:
        self.meldRules.createRule('Concealed Kong', 'FConcealedMinorKong',
                                  explainTemplate='{meldName}', points=16)
        self.meldRules.createRule(
            'Concealed Kong of Terminals', 'FConcealedTerminalsKong',
            explainTemplate='{meldName}', points=32)
        self.meldRules.createRule(
            'Concealed Kong of Honors', 'FConcealedHonorsKong',
            explainTemplate='{meldName}', points=32)

        self.meldRules.createRule('Concealed Pung', 'FConcealedMinorPung',
                                  explainTemplate='{meldName}', points=4)
        self.meldRules.createRule(
            'Concealed Pung of Terminals', 'FConcealedTerminalsPung',
            explainTemplate='{meldName}', points=8)
        self.meldRules.createRule(
            'Concealed Pung of Honors', 'FConcealedHonorsPung',
            explainTemplate='{meldName}', points=8)

        self.meldRules.createRule('Pair of Own Wind', 'FOwnWindPair',
                                  explainTemplate='Pair of Own Wind ({value})', points=2)
        self.meldRules.createRule('Pair of Round Wind', 'FRoundWindPair',
                                  explainTemplate='Pair of Round Wind ({value})', points=2)
        self.meldRules.createRule('Pair of Dragons', 'FDragonPair',
                                  explainTemplate='{meldName}', points=2)

        # bonus tiles:
        self.meldRules.createRule('Flower', 'FFlower',
                                  explainTemplate='{meldName}', points=4)
        self.meldRules.createRule('Season', 'FSeason',
                                  explainTemplate='{meldName}', points=4)


class ClassicalChineseDMJL(ClassicalChinese):

    """classical chinese rules, German rules"""

    def __init__(self, name:Optional[str]=None) ->None:
        super().__init__(name or i18nE('Classical Chinese DMJL'))

    def _initRuleset(self) ->None:
        """set the description"""
        super()._initRuleset()
        self.description = i18n(
            'Classical Chinese as defined by the Deutsche Mah Jongg Liga (DMJL) e.V.')

    def loadRules(self) ->None:
        super().loadRules()
        # the squirming snake is only covered by standard mahjongg rule if
        # tiles are ordered
        self.mjRules.createRule(
            'Squirming Snake', 'FSquirmingSnake', limits=1,
            description=i18n('All tiles of same color. Pung or Kong of 1 and 9, pair of 2, 5 or 8 and two '
                             'Chows of the remaining values'))
        self.handRules.createRule(
            'Little Three Dragons', 'FLittleThreeDragons', doubles=1,
            description=i18n('2 Pungs or Kongs of dragons and 1 pair of dragons'))
        self.handRules.createRule(
            'Big Three Dragons', 'FBigThreeDragons', doubles=2,
            description=i18n('3 Pungs or Kongs of dragons'))
        self.handRules.createRule(
            'Little Four Joys', 'FLittleFourJoys', doubles=1,
            description=i18n('3 Pungs or Kongs of winds and 1 pair of winds'))
        self.handRules.createRule('Big Four Joys', 'FBigFourJoys', doubles=2,
                                  description=i18n('4 Pungs or Kongs of winds'))

        self.winnerRules['OnlyHonors'].doubles = 2  # type:ignore[attr-defined]

        self.penaltyRules.createRule(
            'False Naming of Discard, Claimed for Chow',
            points=-50)
        self.penaltyRules.createRule(
            'False Naming of Discard, Claimed for Pung/Kong',
            points=-100)
        self.penaltyRules.createRule(
            'False Declaration of Mah Jongg by One Player',
            'Oabsolute payees=3', points=-300)
        self.penaltyRules.createRule(
            'False Declaration of Mah Jongg by Two Players',
            'Oabsolute payers=2 payees=2', points=-300)
        self.penaltyRules.createRule(
            'False Declaration of Mah Jongg by Three Players',
            'Oabsolute payers=3', points=-300)
        self.penaltyRules.createRule(
            'False Naming of Discard, Claimed for Mah Jongg',
            'Oabsolute payees=3', points=-300)


class ClassicalChineseBMJA(ClassicalChinese):

    """classical chinese rules, British rules"""

    def __init__(self, name:Optional[str]=None) ->None:
        super().__init__(name or i18nE('Classical Chinese BMJA'))

    def _initRuleset(self) ->None:
        """set the description"""
        super()._initRuleset()
        self.description = i18n(
            'Classical Chinese as defined by the British Mah-Jong Association')

    def addParameterRules(self) ->None:
        """those differ for BMJA from standard"""
        super().addParameterRules()
        self.parameterRules['kongBoxSize'].parameter = 14  # type:ignore[attr-defined]
        self.parameterRules['maxChows'].parameter = 1  # type:ignore[attr-defined]
        self.parameterRules['limit'].parameter = 1000  # type:ignore[attr-defined]
        self.parameterRules['mustDeclareCallingHand'].parameter = True  # type:ignore[attr-defined]

    def loadRules(self) ->None:
        super().loadRules()
        del self.winnerRules['ZeroPointHand']
        originalCall = self.winnerRules.pop('MahJonggwithOriginalCall')
        self.winnerRules.createRule(
            'Original Call', originalCall.definition, doubles=1,
            description=originalCall.description)
        del self.mjRules['NineGates']
        self.mjRules.createRule(
            'Gates of Heaven', 'FGatesOfHeaven', limits=1,
            description=i18n('All tiles concealed of same color: Values 1112345678999'
                             ' with one pair from 2 to 8 (last tile from wall or discarded)'))
        self.mjRules.createRule(
            'Wriggling Snake', 'FWrigglingSnake', limits=1,
            description=i18n('Pair of 1s and a run from 2 to 9 in the same suit with each of the winds'))
        self.mjRules.createRule(
            'Triple Knitting', 'FTripleKnitting', limits=0.5,
            description=i18n('Four sets of three tiles in the different suits and a pair: No Winds or Dragons'))
        self.mjRules.createRule('Knitting', 'FKnitting', limits=0.5,
                                description=i18n('7 pairs of tiles in any 2 out of 3 suits; no Winds or Dragons'))
        self.mjRules.createRule(
            'All pair honors', 'FAllPairHonors', limits=0.5,
            description=i18n('7 pairs of 1s/9s/Winds/Dragons'))
        del self.handRules['OwnFlowerandOwnSeason']
        del self.handRules['ThreeConcealedPongs']
        self.meldRules.createRule('Own Flower', 'FOwnFlower', doubles=1)
        self.meldRules.createRule('Own Season', 'FOwnSeason', doubles=1)
        del self.winnerRules['LastTileTakenfromDeadWall']
        del self.winnerRules['HiddenTreasure']
        del self.winnerRules['FalseColorGame']
        del self.winnerRules['ConcealedTrueColorGame']
        del self.winnerRules['Eastwonninetimesinarow']
        del self.winnerRules['LastTileCompletesPairof28']
        del self.winnerRules['LastTileCompletesPairofTerminalsorHonors']
        del self.winnerRules['LastTileisOnlyPossibleTile']
        del self.winnerRules['TrueColorGame']
        del self.winnerRules['ThreeGreatScholars']
        self.winnerRules.createRule(
            'Buried Treasure', 'FBuriedTreasure', limits=1,
            description=i18n('Concealed pungs of one suit with winds/dragons and a pair'))
        self.winnerRules.createRule('Purity', 'FPurity', doubles=3,
                                    description=i18n('Only same-colored tiles (no chows, dragons or winds)'))
        self.winnerRules.createRule(
            'Three Great Scholars', 'FThreeGreatScholars||Onochow', limits=1,
            description=i18n('3 Pungs or Kongs of dragons plus any pung/kong and a pair'))
        orphans = self.mjRules.pop('ThirteenOrphans')
        self.mjRules.createRule(
            'The 13 Unique Wonders',
            orphans.definition,
            limits=1,
            description=orphans.description)
        self.handRules['AllFlowers'].score.doubles = 2
        self.handRules['AllSeasons'].score.doubles = 2
        self.penaltyRules.createRule(
            'False Naming of Discard, Claimed for Chow/Pung/Kong',
            points=-50)
        self.penaltyRules.createRule(
            'False Declaration of Mah Jongg by One Player',
            'Oabsolute payees=3', limits=-0.5)
        self.winnerRules.createRule(
            'False Naming of Discard, Claimed for Mah Jongg',
            'FFalseDiscardForMJ||Opayforall')

        self.loserRules.createRule(
            'Calling for Only Honors',
            'FCallingHand||Ohand=OnlyHonors',
            limits=0.4)
        self.loserRules.createRule(
            'Calling for Wriggling Snake',
            'FCallingHand||Ohand=WrigglingSnake',
            limits=0.4)
        self.loserRules.createRule(
            'Calling for Triple Knitting',
            'FCallingHand||Ohand=TripleKnitting',
            limits=0.2)
        self.loserRules.createRule(
            'Calling for Gates of Heaven', 'FCallingHand||Ohand=GatesofHeaven',
            limits=0.4)
        self.loserRules.createRule(
            'Calling for Knitting',
            'FCallingHand||Ohand=Knitting',
            limits=0.2)
        self.loserRules.createRule(
            'Calling for Imperial Jade',
            'FCallingHand||Ohand=ImperialJade',
            limits=0.4)
        self.loserRules.createRule(
            'Calling for The 13 Unique Wonders', 'FCallingHand||Ohand=The13UniqueWonders',
            limits=0.4)
        self.loserRules.createRule(
            'Calling for Three Great Scholars', 'FCallingHand||Ohand=ThreeGreatScholars',
            limits=0.4)
        self.loserRules.createRule(
            'Calling for All pair honors',
            'FCallingHand||Ohand=Allpairhonors',
            limits=0.2)
        self.loserRules.createRule(
            'Calling for Heads and Tails',
            'FCallingHand||Ohand=HeadsandTails',
            limits=0.4)
        self.loserRules.createRule(
            'Calling for Four Blessings Hovering over the Door',
            'FCallingHand||Ohand=FourBlessingsHoveringovertheDoor', limits=0.4)
        self.loserRules.createRule(
            'Calling for Buried Treasure',
            'FCallingHand||Ohand=BuriedTreasure',
            limits=0.4)
        self.loserRules.createRule(
            'Calling for Fourfold Plenty',
            'FCallingHand||Ohand=FourfoldPlenty',
            limits=0.4)
        self.loserRules.createRule(
            'Calling for Purity',
            'FCallingHand||Ohand=Purity',
            doubles=3)


class Taiwanese16(ClassicalChinese):

    """Taiwanese rules with 16 tiles in hand"""

    def __init__(self, name:Optional[str]=None) ->None:
        super().__init__(name or i18nE('Taiwanese 16-tile game, FOR DEVELOPERS'))

    def _initRuleset(self) ->None:
        """set the description"""
        super()._initRuleset()
        self.description = i18n(
            'Taiwanese 16-Tiles Mah Jongg, early alpha. See https://bugs.kde.org/show_bug.cgi?id=298821')

    def addParameterRules(self) ->None:
        """those differ for BMJA from standard"""
        super().addParameterRules()
        self.parameterRules['dealtTiles'].parameter = 16  # type:ignore[attr-defined]

    def loadRules(self) ->None:
        super().loadRules()
        del self.mjRules['StandardMahJongg']
        self.mjRules.createRule(
            'Standard Mah Jongg',
            'FStandardMahJongg16',
            points=20)
        del self.winnerRules['ZeroPointHand']
        originalCall = self.winnerRules.pop('MahJonggwithOriginalCall')
        self.winnerRules.createRule(
            'Original Call', originalCall.definition, doubles=1,
            description=originalCall.description)
        del self.handRules['OwnFlowerandOwnSeason']
        del self.handRules['ThreeConcealedPongs']
        self.meldRules.createRule('Own Flower', 'FOwnFlower', doubles=1)
        self.meldRules.createRule('Own Season', 'FOwnSeason', doubles=1)
        del self.winnerRules['LastTileTakenfromDeadWall']
        del self.winnerRules['HiddenTreasure']
        del self.winnerRules['FalseColorGame']
        del self.winnerRules['ConcealedTrueColorGame']
        del self.winnerRules['Eastwonninetimesinarow']
        del self.winnerRules['LastTileCompletesPairof28']
        del self.winnerRules['LastTileCompletesPairofTerminalsorHonors']
        del self.winnerRules['LastTileisOnlyPossibleTile']
        del self.winnerRules['TrueColorGame']
        del self.winnerRules['ThreeGreatScholars']
        self.winnerRules.createRule('Purity', 'FPurity', doubles=3,
                                    description=i18n('Only same-colored tiles (no chows, dragons or winds)'))
        self.handRules['AllFlowers'].score.doubles = 2
        self.handRules['AllSeasons'].score.doubles = 2
        self.penaltyRules.createRule(
            'False Naming of Discard, Claimed for Chow/Pung/Kong',
            points=-50)
        self.penaltyRules.createRule(
            'False Declaration of Mah Jongg by One Player',
            'Oabsolute payees=3', limits=-0.5)
        self.winnerRules.createRule(
            'False Naming of Discard, Claimed for Mah Jongg',
            'FFalseDiscardForMJ||Opayforall')


def load() ->None:
    """load predefined rulesets.
    add new predefined rulesets here.
    """
    assert not PredefinedRuleset.classes
    PredefinedRuleset.classes.add(ClassicalChineseDMJL)
    PredefinedRuleset.classes.add(ClassicalChineseBMJA)
    PredefinedRuleset.classes.add(Taiwanese16)



# -*- coding: utf-8 -*-

"""
Copyright (C) 2010-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

import datetime
from typing import Dict, Tuple, Optional, TYPE_CHECKING, List, Any, Union, cast

from log import logWarning, logException, logDebug, logFailure
from mi18n import i18n, i18nc, i18ncE
from sound import Voice
from tile import Tile, TileTuple, Meld, MeldList
from common import Internal, Debug, Options, ReprMixin
from wind import Wind
from dialogs import Sorry

if TYPE_CHECKING:
    from servertable import ServerTable
    from twisted.internet.defer import Deferred
    from move import Move
    from client import Client
    from player import PlayingPlayer
    from deferredutil import Request
    from user import User
    from humanclient import HumanClient, DlgButton
    from scene import PlayingScene


class Message:

    """those are the message types between client and server. They have no state
    i.e. they never hold real attributes. They only describe the message and actions upon it"""

    defined : Dict[str, 'Message'] = {}
    sendScore = False

# only for mypy:
    Abort:'Message'
    Pung:'ServerMessage'
    Kong:'ServerMessage'
    Chow:'ServerMessage'
    Bonus:'Message'
    MahJongg:'ServerMessage'
    OriginalCall:'ServerMessage'
    Discard:Union['ServerMessage', 'ClientMessage']
    ProposeGameId:'Message'
    TableChanged:'Message'
    ReadyForGameStart:'ServerMessage'
    NoGameStart:'ServerMessage'
    ReadyForHandStart:'ServerMessage'
    InitHand:'ServerMessage'
    SetConcealedTiles:'Message'
    ShowConcealedTiles:'ServerMessage'
    SaveHand:'ServerMessage'
    AskForClaims:'ServerMessage'
    PickedTile:'ServerMessage'
    ActivePlayer:'ServerMessage'
    ViolatesOriginalCall:'Message'
    VoiceId:'Message'
    VoiceData:'Message'
    AssignVoices:'Message'
    ClientWantsVoiceData:'Message'
    ServerWantsVoiceData:'Message'
    ServerGetsVoiceData:'ServerMessage'
    DeclaredKong:'ServerMessage'
    RobbedTheKong:'Message'
    Calling:'Message'
    DangerousGame:'Message'
    NoChoice:'Message'
    UsedDangerousFrom:'Message'
    Draw:'Message'
    Error:'Message'
    NO:'Message'
    OK:'ClientMessage'
    NoClaim:'Message'



    def __init__(self, name:Optional[str]=None, shortcut:Optional[str]=None) ->None:
        """those are the english values"""
        self.name:str = name or self.__class__.__name__.replace('Message', '')
        self.__i18nName:Optional[str] = None
        self.shortcut = shortcut
        # do not use a numerical value because that could easier
        # change with software updates
        className = self.__class__.__name__.replace('Message', '')
        msgName = self.name.replace(' ', '')
        assert className == msgName, f'{className} != {msgName}'

    @property
    def i18nName(self) ->str:
        """only translate when needed - most messages never need this"""
        if self.__i18nName is None:
            self.__i18nName = i18nc('kajongg', self.name)
        return self.__i18nName

    def __str__(self) ->str:
        return self.name

    def __repr__(self) ->str:
        return f'Message.{self.name}'

    def __lt__(self, other:object) ->bool:
        return self.__class__.__name__ < other.__class__.__name__

    @staticmethod
    def jelly(key:Any, value:Any) ->Any:
        """serialize value for wire transfer. The twisted.pb mechanism with
        pb.Copyable is too much overhead"""
        # pylint: disable=too-many-return-statements
        cls = value.__class__
        if cls in (Tile, Meld, MeldList, TileTuple):
            return str(value)
        if isinstance(value, Wind):
            return str(value)
        if isinstance(value, Message):
            return value.name
        if isinstance(value, (list, tuple)):
            if isinstance(value, tuple) and isinstance(value[0], Message):
                if value[1] is None or value[1] == []:
                    return value[0].name
            return type(value)([Message.jelly(key, x) for x in value])
        if isinstance(value, dict):
            return {Message.jelly('key', x): Message.jelly('value', y) for x, y in value.items()}
        if not isinstance(value, (int, bytes, str, float, type(None))):
            raise TypeError(
                f'callRemote got illegal arg: {key} {type(value)}({str(value)})')
        return value

    @staticmethod
    def jellyAll(args:Any, kwargs:Any) ->Tuple[Any, Any]:
        """serialize args and kwargs for wire transfer. The twisted.pb mechanism with
        pb.Copyable is too much overhead"""
        args2 = Message.jelly('args', args)
        kwargs2 = {}
        for key, value in kwargs.items():
            kwargs2[key] = Message.jelly(key, value)
        return args2, kwargs2


class ServerMessage(Message):

    """those classes are used for messages from server to client"""
    # if sendScore is True, this message will send info about player scoring,
    # so the clients can compare
    sendScore = False
    needsGame = True   # message only applies to an existing game

    def clientAction(self, client:'Client', move:'Move') ->Any: # pylint: disable=unused-argument
        """default client action: none - this is a virtual class"""
        logException(
            f'clientAction is not defined for {self}. msg:{move}')

    def serverAction(self, table:'ServerTable', msg:'Request') ->None:  # pylint: disable=unused-argument
        """the server mirrors that and tells all others"""
        logException(f'serverAction is not defined for msg:{msg}')


class ClientMessage(Message):

    """those classes are used for messages from client to server"""

    def buttonCaption(self) ->str:
        """localized, with a & for the shortcut"""
        assert self.shortcut
        i18nShortcut = i18nc(
            'kajongg game dialog:Key for ' +
            self.name,
            self.shortcut)
        return self.i18nName.replace(i18nShortcut, '&' + i18nShortcut, 1)

    def toolTip(self, button:'DlgButton', tile:Tile) ->Tuple[str, bool, str]: # pylint: disable=unused-argument
        """return text and warning flag for button and text for tile for button and text for tile"""
        txt = f'toolTip is not defined for {self.name}'
        logWarning(txt)
        return txt, True, ''

    def serverAction(self, table:'ServerTable',  msg:'Request') ->None: # pylint: disable=unused-argument
        """default server action: none - this is a virtual class"""
        logException(
            f'serverAction is not defined for {self}. msg:{msg}')

    @staticmethod
    def isActivePlayer(table:'ServerTable',  msg:'Request') ->bool:
        """helper: does the message come from the active player?"""
        if table.game and msg.player == table.game.activePlayer:
            return True
        errMsg = f'{msg.player} said {msg.answer} but is not the active player'
        table.abort(errMsg)
        return False


class MessageAbort(ClientMessage):

    """If a client aborts, the server will set the answer for all open requests
    to Message.AbortMessage"""


class NotifyAtOnceMessage(ClientMessage):

    """those classes are for messages from clients that might have to be relayed to the
    other clients right away.

    Example: If a client says Pung, it sends Message.Pung with the flag 'notifying=True'.
    This is relayed to the other 3 clients, helping them in their thinking. When the
    server decides that the Pung is actually to be executed, it sends Message.Pung
    to all 4 clients, but without 'notifying=True'"""

    sendScore = False

    def notifyAction(self, client:'Client', move:'Move') ->Any: # pylint: disable=unused-argument
        """the default action for immediate notifications"""
        assert move.player
        move.player.popupMsg(self)

    @classmethod
    def receivers(cls, request:'Request') ->List['PlayingPlayer']:
        """who should get the notification? Default is all players except the
        player who triggered us"""
        # default: tell all except the source of the notification
        game = request.block.table.game
        if game:
            return [x for x in game.players if x != request.player]
        return []


class PungChowMessage(NotifyAtOnceMessage):

    """common code for Pung and Chow"""

    def __init__(self, name:Optional[str]=None, shortcut:Optional[str]=None) ->None:
        super().__init__(name=name, shortcut=shortcut)

    def toolTip(self, button:'DlgButton', tile:Tile) ->Tuple[str, bool, str]:
        """for the action button which will send this message"""
        if not button.client.game:
            return '', False, ''
        myself = cast('PlayingPlayer', button.client.game.myself)
        maySay = myself.sayable[self]
        if not maySay or not myself.game:
            return '', False, ''
        txt = []
        warn = False
        if myself.originalCall and myself.mayWin:
            warn = True
            txt.append(i18n('saying %1 violates Original Call',
                            self.i18nName))
        dangerousMelds = myself.maybeDangerous(self)
        if dangerousMelds:
            lastDiscard = myself.game.lastDiscard
            warn = True
            if Debug.dangerousGame:
                if not isinstance(maySay, bool) and len(dangerousMelds) != len(maySay):
                    button.client.game.debug(
                        f'only some claimable melds are dangerous: {dangerousMelds}')
            if lastDiscard:
                if len(dangerousMelds) == 1:
                    txt.append(i18n(
                        'claiming %1 is dangerous because you will have to discard a dangerous tile',
                        lastDiscard.name()))
                else:
                    for meld in dangerousMelds:
                        txt.append(i18n(
                            'claiming %1 for %2 is dangerous because you will have to discard a dangerous tile',
                            lastDiscard.name(), str(meld)))
        if not txt:
            txt = [i18n('You may say %1', self.i18nName)]
        return '<br><br>'.join(txt), warn, ''


class MessagePung(PungChowMessage, ServerMessage):

    """somebody said pung and gets the tile"""

    def __init__(self) ->None:
        super().__init__(name=i18ncE('kajongg', 'Pung'),
                         shortcut=i18ncE('kajongg game dialog:Key for Pung', 'P'))

    def serverAction(self, table:'ServerTable',  msg:'Request') ->None:
        """the server mirrors that and tells all others"""
        assert msg.player
        assert isinstance(msg.args, list)
        table.claimTile(msg.player, self, Meld(msg.args[0]), Message.Pung)

    def clientAction(self, client:'Client', move:'Move') ->Any:
        """mirror pung call"""
        return client.claimed(move)


class MessageKong(NotifyAtOnceMessage, ServerMessage):

    """somebody said kong and gets the tile"""

    def __init__(self) ->None:
        super().__init__(name=i18ncE('kajongg', 'Kong'),
                         shortcut=i18ncE('kajongg game dialog:Key for Kong', 'K'))

    def serverAction(self, table:'ServerTable',  msg:'Request') ->None:
        """the server mirrors that and tells all others"""
        assert table.game
        assert msg.player
        assert isinstance(msg.args, list)
        if table.game.lastDiscard:
            table.claimTile(msg.player, self, Meld(msg.args[0]), Message.Kong)
        else:
            table.declareKong(msg.player, Meld(msg.args[0]))

    def toolTip(self, button:'DlgButton', tile:Tile) ->Tuple[str, bool, str]:
        """for the action button which will send this message"""
        if not button.client.game:
            return '', False, ''
        myself = cast('PlayingPlayer', button.client.game.myself)
        maySay = myself.sayable[self]
        if not maySay:
            return '', False, ''
        txt = []
        warn = False
        if myself.originalCall and myself.mayWin:
            warn = True
            assert isinstance(maySay, MeldList)
            txt.append(
                i18n('saying Kong for %1 violates Original Call',
                     Tile(maySay[0][0]).name()))
        if not txt:
            assert isinstance(maySay, MeldList)
            txt = [i18n('You may say Kong for %1', Tile(maySay[0][0]).name())]
        return '<br><br>'.join(txt), warn, ''

    def clientAction(self, client:'Client', move:'Move') ->Any:
        """mirror kong call"""
        assert client.game
        return client.claimed(move) if client.game.lastDiscard else client.declared(move)


class MessageChow(PungChowMessage, ServerMessage):

    """somebody said chow and gets the tile"""

    def __init__(self) ->None:
        super().__init__(name=i18ncE('kajongg', 'Chow'),
                         shortcut=i18ncE('kajongg game dialog:Key for Chow', 'C'))

    def serverAction(self, table:'ServerTable',  msg:'Request') ->None:
        """the server mirrors that and tells all others"""
        assert table.game
        assert isinstance(msg.args, list)
        if table.game.nextPlayer() != msg.player:
            table.abort(
                f'player {msg.player} illegally said Chow, only {table.game.nextPlayer()} may')
        else:
            table.claimTile(msg.player, self, Meld(msg.args[0]), Message.Chow)

    def clientAction(self, client:'Client', move:'Move') ->Any:
        """mirror chow call"""
        return client.claimed(move)


class MessageBonus(ClientMessage):

    """the client says he got a bonus"""

    def serverAction(self, table:'ServerTable', msg:'Request') ->None:
        """the server mirrors that"""
        if self.isActivePlayer(table, msg):
            table.pickTile()


class MessageMahJongg(NotifyAtOnceMessage, ServerMessage):

    """somebody sayd mah jongg and wins"""
    sendScore = True

    def __init__(self) ->None:
        super().__init__(name=i18ncE('kajongg', 'Mah Jongg'),
                         shortcut=i18ncE('kajongg game dialog:Key for Mah Jongg', 'M'))

    def serverAction(self, table:'ServerTable',  msg:'Request') ->None:
        """the server mirrors that and tells all others"""
        table.claimMahJongg(msg)

    def toolTip(self, button:'DlgButton', tile:Tile) ->Tuple[str, bool, str]:
        """return text and warning flag for button and text for tile"""
        return i18n('Press here and you win'), False, ''

    def clientAction(self, client:'Client', move:'Move') ->Any:
        """mirror the mahjongg action locally. Check if the balances are correct."""
        assert move.player
        return move.player.declaredMahJongg(move.melds, move.withDiscardTile,
                                            move.lastTile, move.lastMeld)


class MessageOriginalCall(NotifyAtOnceMessage, ServerMessage):

    """somebody made an original call"""

    def __init__(self) ->None:
        super().__init__(name=i18ncE('kajongg', 'Original Call'),
                         shortcut=i18ncE('kajongg game dialog:Key for Original Call', 'O'))

    def serverAction(self, table:'ServerTable',  msg:'Request') ->None:
        """the server tells all others"""
        table.clientDiscarded(msg)

    def toolTip(self, button:'DlgButton', tile:Tile) ->Tuple[str, bool, str]:
        """for the action button which will send this message"""
        # TODO: factor out common assertions
        assert isinstance(tile, Tile), tile
        if not button.client.game:
            return '', False, ''
        myself = button.client.game.myself
        isCalling = bool((myself.hand - tile).callingHands)
        if not isCalling:
            txt = i18n(
                'discarding %1 and declaring Original Call makes this hand unwinnable',
                tile.name())
            return txt, True, txt
        return (i18n(
            'Discard a tile, declaring Original Call meaning you need only one '
            'tile to complete the hand and will not alter the hand in any way (except bonus tiles)'),
                False, '')

    def clientAction(self, client:'Client', move:'Move') ->Any:
        """mirror the original call"""
        player = move.player
        assert player
        if client.thatWasMe(player):
            player.originalCallingHand = player.hand
            if Debug.originalCall:
                logDebug(
                    f'{player} gets originalCallingHand:{player.originalCallingHand}')
        player.originalCall = True
        assert client.game
        client.game.addCsvTag('originalCall')
        return client.ask(move, [Message.OK])


class MessageDiscard(ClientMessage, ServerMessage):

    """somebody discarded a tile"""
 #   sendScore = True

    def __init__(self) ->None:
        super().__init__(name=i18ncE('kajongg', 'Discard'),
                         shortcut=i18ncE('kajongg game dialog:Key for Discard', 'D'))

    def serverAction(self, table:'ServerTable',  msg:'Request') ->None:
        """the server mirrors that action"""
        table.clientDiscarded(msg)

    def toolTip(self, button:'DlgButton', tile:Tile) ->Tuple[str, bool, str]:
        """for the action button which will send this message"""
        assert isinstance(tile, Tile), tile
        if not button.client.game:
            return '', False, ''
        game = button.client.game
        myself = cast('PlayingPlayer', game.myself)
        txt = []
        warn = False
        if myself.violatesOriginalCall(tile):
            txt.append(
                i18n('discarding %1 violates Original Call', tile.name()))
            warn = True
        if game.dangerousFor(myself, tile):
            txt.append(i18n('discarding %1 is Dangerous Game', tile.name()))
            warn = True
        if not txt:
            txt = [i18n('discard the least useful tile')]
        txtStr = '<br><br>'.join(txt)
        return txtStr, warn, txtStr

    def clientAction(self, client:'Client', move:'Move') ->Any:
        """execute the discard locally"""
        assert move.player
        if client.isHumanClient() and Internal.scene:
            assert move.player.handBoard
            move.player.handBoard.setEnabled(False)
        move.player.speak(move.tile.name2())
        assert client.game
        return client.game.hasDiscarded(move.player, move.tile)


class MessageProposeGameId(ServerMessage):

    """the game server proposes a new game id. We check if it is available
    in our local data base - we want to use the same gameid everywhere"""
    needsGame = False

    def clientAction(self, client:'Client', move:'Move') ->Any:
        """ask the client"""
        # move.playerNames are the players in seating order
        # we cannot just use table.playerNames - the seating order is now
        # different (random)
        assert move.gameid
        return client.reserveGameId(move.gameid)


class MessageTableChanged(ServerMessage):

    """somebody joined or left a table"""
    needsGame = False

    def clientAction(self, client:'Client', move:'Move') ->Any:
        """update our copy"""
        client.tableChanged(move.source)


class MessageReadyForGameStart(ServerMessage):

    """the game server asks us if we are ready for game start"""
    needsGame = False

    def clientAction(self, client:'Client', move:'Move') ->Any:
        """ask the client"""
        def hideTableList(result:Any) ->Any:
            """hide it only after player says I am ready"""
            # set scene.game first, otherwise tableList.hide()
            # sees no current game and logs out
            if result == Message.OK:
                if client.game and Internal.scene:
                    Internal.scene.game = client.game
            if result == Message.OK and client.tableList and client.tableList.isVisible():
                if Debug.table:
                    logDebug(
                        f'{client.name} hiding table list because game started')
                client.tableList.hide()
            return result
        # move.playerNames are the players in seating order
        # we cannot just use table.playerNames - the seating order is now
        # different (random)
        assert move.tableid
        assert move.gameid
        return client.readyForGameStart(  # type:ignore[attr-defined]
            move.tableid, move.gameid,
            move.wantedGame, move.playerNames, shouldSave=move.shouldSave).addCallback(
                hideTableList).addErrback(logFailure)


class MessageNoGameStart(NotifyAtOnceMessage):

    """the client says he does not want to start the game after all"""
    needsGame = False

    def notifyAction(self, client:'Client', move:'Move') ->Any:
        assert move.player
        client = cast('HumanClient', client)
        if client.beginQuestion or client.game:
            Sorry(i18n('%1 is not ready to start the game', move.player.name))
        if client.beginQuestion:
            client.beginQuestion.cancel()
        elif client.game:
            return client.game.close()
        return None

    @classmethod
    def receivers(cls, request:'Request') ->List['PlayingPlayer']:
        """notification is not needed for those who already said no game"""
        result = [x.player for x in request.block.requests if x.answer != Message.NoGameStart]
        return cast(List['PlayingPlayer'], result)


class MessageReadyForHandStart(ServerMessage):

    """the game server asks us if we are ready for a new hand"""

    def clientAction(self, client:'Client', move:'Move') ->Any:
        """ask the client"""
        return client.readyForHandStart(move.playerNames, move.mustRotateWinds)


class MessageInitHand(ServerMessage):

    """the game server tells us to prepare a new hand"""

    def clientAction(self, client:'Client', move:'Move') ->Any:
        """prepare a new hand"""
        assert client.game
        assert client.game.wall
        client.game.divideAt = move.divideAt
        client.game.wall.divide()
        if hasattr(client, 'shutdownHumanClients'):
            client.shutdownHumanClients(exception=client)
        scene = cast('PlayingScene', Internal.scene)
        if scene:
            scene.mainWindow.setWindowTitle(
                i18n(
                    'Kajongg <numid>%1</numid>',
                    client.game.point.seed))
            scene.discardBoard.setRandomPlaces(client.game)
        client.game.initHand()


class MessageSetConcealedTiles(ServerMessage):

    """the game server assigns tiles to player"""

    def clientAction(self, client:'Client', move:'Move') ->Any:
        """set tiles for player"""
        assert move.player
        assert client.game
        assert client.game.wall
        return move.player.addConcealedTiles(client.game.wall.deal(move.tiles), animated=False)  # type:ignore[arg-type]


class MessageShowConcealedTiles(ServerMessage):

    """the game server assigns tiles to player"""

    def clientAction(self, client:'Client', move:'Move') ->Any:
        """set tiles for player"""
        assert move.player
        return move.player.showConcealedTiles(move.tiles, move.show)


class MessageSaveHand(ServerMessage):

    """the game server tells us to save the hand"""

    def clientAction(self, client:'Client', move:'Move') ->Any:
        """save the hand"""
        assert client.game
        return client.game.saveHand()


class MessageAskForClaims(ServerMessage):

    """the game server asks us if we want to claim a tile"""

    def clientAction(self, client:'Client', move:'Move') ->Any:
        """ask the player"""
        assert move.player
        if client.thatWasMe(move.player):
            raise ValueError(
                f'Server asked me({move.player}) for claims but I just discarded that tile!')
        _ = (Message.NoClaim, Message.Chow, Message.Pung, Message.Kong, Message.MahJongg)
        choice = list(cast(ClientMessage, x) for x in _)
        return client.ask(move, choice)


class MessagePickedTile(ServerMessage):

    """the game server tells us who picked a tile"""

    def clientAction(self, client:'Client', move:'Move') ->Any:
        """mirror the picked tile"""
        assert move.player
        assert move.player.pickedTile(move.deadEnd, tileName=move.tile) == move.tile, \
            (move.player.lastTile, move.tile)
        if client.thatWasMe(move.player):
            return (Message.Bonus, move.tile) if move.tile.isBonus else client.myAction(move)
        return None


class MessageActivePlayer(ServerMessage):

    """the game server tells us whose turn it is"""

    def clientAction(self, client:'Client', move:'Move') ->Any:
        """set the active player"""
        assert client.game
        assert move.player
        client.game.activePlayer = move.player


class MessageViolatesOriginalCall(ServerMessage):

    """the game server tells us who violated an original call"""

    def __init__(self) ->None:
        ServerMessage.__init__(
            self,
            name=i18ncE('kajongg',
                        'Violates Original Call'))

    def clientAction(self, client:'Client', move:'Move') ->Any:
        """violation: player may not say mah jongg"""
        assert move.player
        move.player.popupMsg(self)
        move.player.mayWin = False
        if Debug.originalCall:
            logDebug(f'{move.player}: cleared mayWin')
        return client.ask(move, [Message.OK])


class MessageVoiceId(ServerMessage):

    """we got a voice id from the server. If we have no sounds for
    this voice, ask the server"""

    def clientAction(self, client:'Client', move:'Move') ->Any:
        """the server gave us a voice id about another player"""
        assert Internal.Preferences
        if Internal.Preferences.useSounds and Options.gui:
            assert move.player
            move.player.voice = Voice.locate(cast(str, move.source))
            if not move.player.voice:
                return Message.ClientWantsVoiceData, move.source
        return None


class MessageVoiceData(ServerMessage):

    """we got voice sounds from the server, assign them to the player voice"""

    def clientAction(self, client:'Client', move:'Move') ->Any:
        """server sent us voice sounds about somebody else"""
        assert move.md5sum
        assert move.player
        move.player.voice = Voice(move.md5sum, cast(bytes, move.source))
        if Debug.sound:
            logDebug(f'{move.player} gets voice data {move.player.voice} '
                     f'from server, language={move.player.voice.language()}')


class MessageAssignVoices(ServerMessage):

    """The server tells us that we now got all voice data available"""

    def clientAction(self, client:'Client', move:'Move') ->Any:
        assert Internal.Preferences
        if Internal.Preferences.useSounds and Options.gui:
            assert client.game
            client.game.assignVoices()


class MessageClientWantsVoiceData(ClientMessage):

    """This client wants voice sounds"""


class MessageServerWantsVoiceData(ServerMessage):

    """The server wants voice sounds from a client"""

    def clientAction(self, client:'Client', move:'Move') ->Any:
        """send voice sounds as requested to server"""
        assert move.player
        assert move.player.voice
        if Debug.sound:
            logDebug(f'{move.player}: send wanted voice data {move.player.voice} to server')
        return Message.ServerGetsVoiceData, move.player.voice.archiveContent


class MessageServerGetsVoiceData(ClientMessage):

    """The server gets voice sounds from a client"""

    def serverAction(self, table:'ServerTable', msg:'Request') ->None:
        """save voice sounds on the server"""
        assert isinstance(msg.args, list)
        assert msg.player
        voice = msg.player.voice
        assert voice
        voice.archiveContent = msg.args[0]
        if Debug.sound:
            if voice.oggFiles():
                logDebug(f'{msg.player}: server got wanted voice data {voice}')
            else:
                logDebug(f'{msg.player}: server got empty voice data {voice} (arg0={repr(msg.args[0][:100])})')


class MessageDeclaredKong(ServerMessage):

    """the game server tells us who declared a kong"""

    def clientAction(self, client:'Client', move:'Move') ->Any:
        """mirror the action locally"""
        prompts = None
        assert move.meld
        assert move.player
        if not client.thatWasMe(move.player):
            if len(move.meld) != 4 or move.meld[0].isConcealed:
                # do not do this when adding a 4th tile to an exposed pung
                move.player.showConcealedTiles(move.meld)
            else:
                move.player.showConcealedTiles(TileTuple(move.meld[3]))
            prompts = [cast(ClientMessage, Message.NoClaim), cast(ClientMessage, Message.MahJongg)]
        move.exposedMeld = move.player.exposeMeld(move.meld)
        return client.ask(move, prompts) if prompts else None


class MessageRobbedTheKong(NotifyAtOnceMessage, ServerMessage):

    """the game server tells us who robbed the kong"""

    def clientAction(self, client:'Client', move:'Move') ->Any:
        """mirror the action locally"""
        assert client.game
        prevMove = next(client.game.lastMoves(only=[Message.DeclaredKong]))
        assert prevMove.player
        prevMove.player.getsRobbed(prevMove.meld[0].concealed)
        assert move.player
        move.player.robsTile()
        client.game.addCsvTag(
            f'robbedKong{prevMove.meld[1]}',
            forAllPlayers=True)


class MessageCalling(ServerMessage):

    """the game server tells us who announced a calling hand"""

    def clientAction(self, client:'Client', move:'Move') ->Any:
        """tell user and save this information locally"""
        assert move.player
        move.player.popupMsg(self)
        move.player.isCalling = True
        return client.ask(move, [Message.OK])


class MessageDangerousGame(ServerMessage):

    """the game server tells us who played dangerous game"""

    def __init__(self) ->None:
        super().__init__(name=i18ncE('kajongg', 'Dangerous Game'))

    def clientAction(self, client:'Client', move:'Move') ->Any:
        """mirror the dangerous game action locally"""
        assert move.player
        move.player.popupMsg(self)
        move.player.playedDangerous = True
        return client.ask(move, [Message.OK])


class MessageNoChoice(ServerMessage):

    """the game server tells us who had no choice avoiding dangerous game"""

    def __init__(self) ->None:
        super().__init__(name=i18ncE('kajongg', 'No Choice'))
        self.move:Optional['Move'] = None

    def clientAction(self, client:'Client', move:'Move') ->Any:
        """mirror the no choice action locally"""
        self.move = move
        assert move.player
        move.player.popupMsg(self)
        move.player.claimedNoChoice = True
        move.player.showConcealedTiles(move.tiles)
        # otherwise we have a visible artifact of the discarded tile.
        # Only when animations are disabled. Why?
#        Internal.mainWindow.centralView.resizeEvent(None)
        return client.ask(move, [Message.OK]).addCallback(self.hideConcealedAgain).addErrback(logFailure)

    def hideConcealedAgain(self, result:Tuple[Message, None]) ->Tuple[Message, None]:
        """only show them for explaining the 'no choice'"""
        assert self.move
        assert self.move.player
        self.move.player.showConcealedTiles(self.move.tiles, False)
        return result


class MessageUsedDangerousFrom(ServerMessage):

    """the game server tells us somebody claimed a dangerous tile"""

    def clientAction(self, client:'Client', move:'Move') ->Any:
        assert client.game
        fromPlayer = cast('PlayingPlayer', client.game.playerByName(str(move.source)))
        assert move.player
        move.player.usedDangerousFrom = fromPlayer
        if Debug.dangerousGame:
            logDebug(f'{move.player} claimed a dangerous tile discarded by {fromPlayer}')


class MessageDraw(ServerMessage):

    """the game server tells us nobody said mah jongg"""
    sendScore = True


class MessageError(ServerMessage):

    """a client errors"""
    needsGame = False

    def clientAction(self, client:'Client', move:'Move') ->Any:
        """show the error message from server"""
        return logWarning(str(move.source))


class MessageNO(ClientMessage):

    """a client says no"""


class MessageOK(ClientMessage):

    """a client says OK"""

    def __init__(self) ->None:
        super().__init__(name=i18ncE('kajongg', 'OK'),
                         shortcut=i18ncE('kajongg game dialog:Key for OK', 'O'))

    def toolTip(self, button:'DlgButton', tile:Tile) ->Tuple[str, bool, str]:
        """return text and warning flag for button and text for tile for button and text for tile"""
        return i18n('Confirm that you saw the message'), False, ''


class MessageNoClaim(NotifyAtOnceMessage, ServerMessage):

    """A player explicitly says he will not claim a tile"""

    def __init__(self) ->None:
        super().__init__(name=i18ncE('kajongg', 'No Claim'),
                         shortcut=i18ncE('kajongg game dialog:Key for No claim', 'N'))

    def toolTip(self, button:'DlgButton', tile:Tile) ->Tuple[str, bool, str]:
        """return text and warning flag for button and text for tile for button and text for tile"""
        return i18n('You cannot or do not want to claim this tile'), False, ''

    @classmethod
    def receivers(cls, request:'Request') ->List['PlayingPlayer']:
        """no Claim notifications are not needed for those who already answered"""
        result = [x.player for x in request.block.requests if x.answer is None]
        return cast(List['PlayingPlayer'], result)


def __scanSelf() ->None:
    """for every message defined in this module which can actually be used for traffic,
    generate a class variable Message.msg where msg is the name (without spaces)
    of the message. Example: 'Message.NoClaim'.
    Those will be used as stateless constants. Also add them to dict Message.defined, but with spaces."""
    if Message.defined:
        return
    for glob in globals().values():
        if hasattr(glob, "__mro__"):
            if glob.__mro__[-2] == Message and len(glob.__mro__) > 2:
                if glob.__name__.startswith('Message'):
                    try:
                        msg = glob()
                    except Exception:
                        logDebug(f'cannot instantiate {glob.__name__}')
                        raise
                    type.__setattr__(
                        Message, msg.name.replace(' ', ''), msg)
                    Message.defined[msg.name] = msg


class ChatMessage(ReprMixin):

    """holds relevant info about a chat message"""

    def __init__(self, tableid:Union[Tuple[Any, ...], int], fromUser:Optional[str]=None,
                 message:Optional[str]=None, isStatusMessage:bool=False) ->None:
        if isinstance(tableid, tuple):
            self.tableid, hour, minute, second, self.fromUser, self.message, self.isStatusMessage = tableid
            self.timestamp = datetime.time(
                hour=hour,
                minute=minute,
                second=second)
        else:
            self.tableid = tableid
            self.fromUser = fromUser
            self.message = message
            self.isStatusMessage = isStatusMessage
            self.timestamp = datetime.datetime.utcnow().time()

    def localtimestamp(self) ->datetime.datetime:
        """convert from UTC to local"""
        now = datetime.datetime.now()
        utcnow = datetime.datetime.utcnow()
        result = datetime.datetime.combine(
            datetime.date.today(),
            self.timestamp)
        return result + (now - utcnow)

    def __str__(self) ->str:
        local = self.localtimestamp()
        # pylint says something about NotImplemented, check with later versions
        _ = i18n(self.message)
        if self.isStatusMessage:
            _ = f'[{_}]'
        return (f'{int(local.hour):02}:{int(local.minute):02}:{int(local.second):02} '
                f'{self.fromUser}: {i18n(self.message)}')

    def asList(self) ->Tuple[Any, ...]:
        """encode me for network transfer"""
        return (
            self.tableid, self.timestamp.hour, self.timestamp.minute, self.timestamp.second,
            self.fromUser, self.message, self.isStatusMessage)

__scanSelf()



# -*- coding: utf-8 -*-

"""
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

import csv
import subprocess
import datetime

from enum import IntEnum
from functools import total_ordering

from typing import Tuple, List, Any, Optional, Dict

from common import Options, ReprMixin
from player import Player, Players

class CsvWriter:
    """how we want it"""
    def __init__(self, filename:str, mode:str='w') ->None:
        self.outfile = open(filename, mode, encoding='utf-8')  # pylint: disable=consider-using-with
        self.__writer = csv.writer(self.outfile, delimiter=Csv.delimiter)

    def writerow(self, row:'CsvRow') ->None:
        """write one row"""
        self.__writer.writerow([str(cell) for cell in row])  # type:ignore[attr-defined]

    def __del__(self) ->None:
        """clean up"""
        self.outfile.close()


class Csv:
    """how we want it"""

    delimiter = ';'

    @staticmethod
    def reader(filename:str) ->Any:
        """return a generator for decoded strings"""
        # see https://stackoverflow.com/questions/51264355/how-to-type-annotate-object-returned-by-csv-writer
        return csv.reader(open(filename, 'r', encoding='utf-8'), delimiter=Csv.delimiter)

@total_ordering
class CsvRow(ReprMixin):
    """represent a row in kajongg.csv"""

    Fields = IntEnum('Fields', 'RULESET AI COMMIT PY_VERSION GAME TAGS PLAYERS', start=0)

    commitDates : Dict[str, datetime.datetime] = {}

    def __init__(self, row:List[str]) ->None:
        self.row = row
        self.ruleset, self.aiVariant, self.commit, self.py_version, self.game, self.tags = row[:6]
        self.winner = None
        rest = row[6:]
        players:List[Player] = []
        while rest:
            name, balance, wonCount, winner = rest[:4]
            player = Player(None, name)
            player.balance = int(balance)
            player.wonCount = int(wonCount)
            players.append(player)
            if winner:
                self.winner = player
            rest = rest[4:]
        self.players = Players(players)

    @property
    def commitDate(self) ->datetime.datetime:
        """return datetime"""
        if self.commit not in self.commitDates:
            try:
                self.commitDates[self.commit] = datetime.datetime.fromtimestamp(
                    int(subprocess.check_output(
                        f'git show -s --format=%ct {self.commit}'.split(), stderr=subprocess.DEVNULL)))
            except subprocess.CalledProcessError:
                self.commitDates[self.commit] = datetime.datetime.fromtimestamp(0)
        return self.commitDates[self.commit]

    @property
    def game(self) ->str:
        """return the game"""
        return self.row[self.Fields.GAME]

    @game.setter
    def game(self, value:str) ->None:
        self.row[self.Fields.GAME] = value

    @property
    def ruleset(self) ->str:
        """return the ruleset"""
        return self.row[self.Fields.RULESET]

    @ruleset.setter
    def ruleset(self, value:str) ->None:
        self.row[self.Fields.RULESET] = value

    @property
    def aiVariant(self) ->str:
        """return the AI used"""
        return self.row[self.Fields.AI]

    @aiVariant.setter
    def aiVariant(self, value:str) ->None:
        self.row[self.Fields.AI] = value

    @property
    def commit(self) ->str:
        """return the git commit"""
        return self.row[self.Fields.COMMIT]

    @commit.setter
    def commit(self, value:str) ->None:
        self.row[self.Fields.COMMIT] = value

    @property
    def py_version(self) ->str:
        """return the python version"""
        return self.row[self.Fields.PY_VERSION]

    @py_version.setter
    def py_version(self, value:str) ->None:
        self.row[self.Fields.PY_VERSION] = value

    @property
    def tags(self) ->str:
        """return the tags"""
        return self.row[self.Fields.TAGS]

    @tags.setter
    def tags(self, value:str) ->None:
        self.row[self.Fields.TAGS] = value

    def result(self) ->Tuple[str, ...]:
        """return a tuple with the fields holding the result"""
        return tuple(self.row[self.Fields.PLAYERS:])

    def write(self) ->None:
        """write to Options.csv"""
        assert Options.csv
        writer = CsvWriter(Options.csv, mode='a')
        writer.writerow(self.row)
        del writer

    def __eq__(self, other:Any) ->bool:
        return self.row == other.row

    def sortkey(self) ->List[Any]:
        """return string for comparisons"""
        result = [self.game, self.ruleset, self.aiVariant,
                  self.commitDate or datetime.datetime.fromtimestamp(0), self.py_version]
        result.extend(self.row[self.Fields.TAGS:])
        return result

    def __lt__(self, other:Any) ->bool:
        return self.sortkey() < other.sortkey()

    def __getitem__(self, field:int) ->str:
        """direct access to row"""
        return self.row[field]

    def __hash__(self) ->int:
        return hash(tuple(self.row))

    def data(self, field:int) ->str:
        """return a string representing this field for messages"""
        result = self.row[field]
        if field == self.Fields.COMMIT:
            result = f'{result}({self.commitDate})'
        return result

    def differs_for(self, other:'CsvRow') ->Optional[Tuple[str, str]]:
        """return the field names for the source attributes causing a difference.
        Possible values are commit and py_version. If both rows are identical, return None."""
        if self.row[self.Fields.PLAYERS:] != other.row[self.Fields.PLAYERS:]:
            differing = []
            same = []
            for cause in (self.Fields.COMMIT, self.Fields.PY_VERSION):
                if self.row[cause] != other.row[cause]:
                    _ = f'{cause.name} {self.data(cause)} != {other.data(cause)}'
                    differing.append(_)
                else:
                    _ = f'{cause.name} {self.data(cause)}'
                    same.append(_)
            return ', '.join(differing), ', '.join(same)
        return None

    def neutralize(self) ->None:
        """for comparisons"""
        for idx, field in enumerate(self.row):
            field = field.replace(' ', '')
            if field.startswith('Tester ') or field.startswith('Tüster'):
                field = 'Tester'
            if 'MEM' in field:
                parts = field.split(',')
                for part in parts[:]:
                    if part.startswith('MEM'):
                        parts.remove(part)
                field = ','.join(parts)
            self.row[idx] = field

    def __str__(self) ->str:
        return (f'Game {self.game} {self.ruleset} AI={self.aiVariant} '
                f'commit={self.commit}({self.commitDate}) py={self.py_version} {self.tags}')



# -*- coding: utf-8 -*-

"""
Copyright (C) 2009-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only


The DBPasswordChecker is based on an example from the book
Twisted Network Programming Essentials by Abe Fettig, 2006
O'Reilly Media, Inc., ISBN 0-596-10032-9
"""

import datetime
from typing import Optional, TYPE_CHECKING, List, Any, Union

from twisted.internet.defer import fail
from twisted.spread import pb

from common import Internal, Debug, Options, ReprMixin
from servercommon import srvError
from log import logDebug
from mi18n import i18nE
from query import Query

if TYPE_CHECKING:
    from twisted.internet.defer import Deferred
    from server import MJServer
    from rule import Ruleset


class User(pb.Avatar, ReprMixin):

    """the twisted avatar"""

    def __init__(self, userid:str) ->None:
        self.name = Query(
            'select name from player where id=?',
            (userid,
            )).records[0][0]
        self.mind:Optional[pb.RemoteReference] = None
        self.server:Optional['MJServer'] = None
        self.dbIdent:Optional[str] = None
        self.voiceId:Optional[str] = None
        self.maxGameId:Optional[int] = None
        self.lastPing:Optional['datetime.datetime'] = None
        self.pinged()

    def pinged(self) ->None:
        """time of last ping or message from user"""
        self.lastPing = datetime.datetime.now()
        if self.server:
            self.server.lastPing = self.lastPing

    def source(self) ->str:
        """how did he connect?"""
        if self.mind is None:
            result = 'SOURCE UNKNOWN: self.mind is None'
        else:
            result = str(self.mind.broker.transport.getPeer())
        if 'UNIXAddress' in result:
            # socket: we want to get the socket name
            result = Options.socket or 'ERROR: Options.socket is None'
        return result

    def attached(self, mind:pb.RemoteReference) ->None:
        """override pb.Avatar.attached"""
        self.mind = mind
        assert self.server
        self.server.login(self)

    def detached(self, unusedMind:Optional[pb.RemoteReference]=None) ->None:
        """override pb.Avatar.detached"""
        if Debug.connections:
            logDebug(
                f'{self}: connection detached from {self.source()}')
        assert self.server
        self.server.logout(self)
        self.mind = None

    def perspective_setClientProperties(
            self, dbIdent:str, voiceId:str, maxGameId:int, clientVersion:Optional[str]=None) ->Optional['Deferred']:
        """perspective_* methods are to be called remotely"""
        self.pinged()
        self.dbIdent = dbIdent
        self.voiceId = voiceId
        self.maxGameId = maxGameId
        serverVersion = Internal.defaultPort
        if clientVersion != serverVersion:
            if clientVersion is None:
                # client passed no version info
                return fail(srvError(
                                     i18nE(
                                         'Your client has a version older than 4.9.0 but you need %1 for this server'),
                                     serverVersion))
            return fail(srvError(pb.Error,
                                 i18nE(
                                     'Your client has version %1 but you need %2 for this server'),
                                 clientVersion or '<4.9.0',
                                 serverVersion))
        if Debug.table:
            logDebug(f'client has dbIdent={self.dbIdent} voiceId={self.voiceId} '
                     f'maxGameId={self.maxGameId} clientVersion {clientVersion}')
        assert self.server
        self.server.sendTables(self)
        return None

    def perspective_ping(self) ->None:
        """perspective_* methods are to be called remotely"""
        return self.pinged()

    def perspective_needRulesets(self, rulesetHashes:List[str]) ->List[List[List[Union[int, str, float]]]]:
        """perspective_* methods are to be called remotely"""
        assert self.server
        return self.server.needRulesets(rulesetHashes)

    def perspective_joinTable(self, tableid:int) ->bool:
        """perspective_* methods are to be called remotely"""
        assert self.server
        return self.server.joinTable(self, tableid)

    def perspective_leaveTable(self, tableid:int) ->bool:
        """perspective_* methods are to be called remotely"""
        assert self.server
        return self.server.leaveTable(self, tableid, f'correctly left table {tableid}')

    def perspective_newTable(
            self, ruleset:str, playOpen:bool, autoPlay:bool, wantedGame: str,
            tableId:Optional[int]=None) ->Optional[Any]:
        """perspective_* methods are to be called remotely"""
        assert self.server
        return self.server.newTable(self, ruleset, playOpen, autoPlay, wantedGame, tableId)

    def perspective_startGame(self, tableid:int) ->None:
        """perspective_* methods are to be called remotely"""
        assert self.server
        self.server.startGame(self, tableid)

    def perspective_logout(self) ->None:
        """perspective_* methods are to be called remotely"""
        self.detached(None)

    def perspective_chat(self, chatString:str) ->None:
        """perspective_* methods are to be called remotely"""
        self.pinged()
        assert self.server
        self.server.chat(chatString)

    def __str__(self) ->str:
        return self.name



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Copyright (C) 2009-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

from server import kajonggServer
from util import checkMemory

kajonggServer()
checkMemory()
# profileMe()



# -*- coding: utf-8 -*-

"""
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

from typing import Optional, Union, TYPE_CHECKING, Any, Generator, Literal, cast

from twisted.internet.defer import succeed, Deferred

from log import logDebug, logFailure, logError
from mi18n import i18n
from common import LIGHTSOURCES, Internal, isAlive, ZValues, Debug
from common import ReprMixin, Speeds, id4
from wind import Wind, NoWind

from qt import Qt, QMetaObject
from qt import QGraphicsScene, QGraphicsItem, QGraphicsRectItem, QPen, QColor

from dialogs import QuestionYesNo
from guiutil import decorateWindow, sceneRotation
from board import SelectorBoard, DiscardBoard
from tileset import Tileset
from tile import Tile, Meld
from humanclient import HumanClient
from uitile import UITile
from uiwall import UIWall
from animation import AnimationSpeed, afterQueuedAnimations
from scoringdialog import ScoringDialog

if TYPE_CHECKING:
    from qt import QEvent, QKeyEvent, QFocusEvent, QWidget, QPointF
    from game import Game
    from handboard import HandBoard
    from mainwindow import MainWindow
    from humanclient import ClientDialog
    from board import Board

class FocusRect(QGraphicsRectItem, ReprMixin):

    """show a focusRect with blue border around focused tile or meld.
    We can NOT do this as a child of the item or the focus rect would
    have to stay within tile face: The adjacent tile will cover the
    focus rect because Z order is only relevant for items having the
    same parent"""

    def __init__(self) ->None:
        QGraphicsRectItem.__init__(self)
        pen = QPen(QColor('blue'))
        pen.setWidth(6)
        self.setPen(pen)
        self.setZValue(ZValues.markerZ)
        self._board:Optional[Union[SelectorBoard, DiscardBoard]] = None
        self.hide()

    @property
    def board(self) ->Optional[Union[SelectorBoard, DiscardBoard]]:
        """current board the focusrect is on"""
        return self._board

    @board.setter
    def board(self, value: Optional[Union[SelectorBoard, DiscardBoard]]) ->None:
        """assign and show/hide as needed"""
        if value and not isAlive(value):
            logDebug(
                f'assigning focusRect to a non-alive board {type(value)}/{value}')
            return
        if value:
            if Debug.focusable:
                logDebug(f'FocusRect {self}: board: {self._board} -> {value}')
            self._board = value
            self.refresh()

    @afterQueuedAnimations  # type:ignore[arg-type]
    def refresh(self, unusedDeferredResult:Optional['Deferred']=None) ->None:
        """show/hide on correct position after queued animations end"""
        board = self.board
        if not board:
            # for mypy
            return
        if not isAlive(board) or not isAlive(self):
            if isAlive(self):
                self.setVisible(False)
            return
        rect = board.tileFaceRect()
        rect.setWidth(rect.width() * board.focusRectWidth())
        self.setRect(rect)
        self.setRotation(sceneRotation(board))
        self.setScale(board.scale())
        if board.focusTile:
            board.focusTile.setFocus()
            self.setPos(cast('QPointF', board.focusTile.pos))
        assert Internal.scene
        game = Internal.scene.game
        if game is None:
            self.setVisible(False)
        else:
            self.setVisible(board.isVisible() and bool(board.focusTile)
                        and board.isEnabled() and board.hasLogicalFocus and not game.autoPlay)

    def __str__(self) ->str:
        """for debugging"""
        return f"FocusRect_{id4(self)}({self.board if self.board else 'NOBOARD'})"

class SceneWithFocusRect(QGraphicsScene):

    """our scene with a potential Qt bug fix. FocusRect is a blue frame around a tile or meld"""

    def __init__(self) ->None:
        QGraphicsScene.__init__(self)
        self.focusRect = FocusRect()
        self.addItem(self.focusRect)

    @property
    def focusBoard(self) ->Optional['Board']:
        """get / set the board that has its focusRect shown"""
        return self.focusRect.board

    @focusBoard.setter
    def focusBoard(self, board:Optional[Union[DiscardBoard, SelectorBoard]]) ->None:
        """get / set the board that has its focusRect shown"""
        if Debug.focusable:
            logDebug(f'scene.focusRect gets focusBoard {board}, has focusItem {self.focusItem()}')
        self.focusRect.board = board
        if Debug.focusable:
            logDebug(f'scene.focusRect is set, scene has focusItem {self.focusItem()}')


class GameScene(SceneWithFocusRect):

    """the game field"""

    def __init__(self, parent:Optional['QWidget']=None) ->None:
        assert parent
        self.mainWindow = cast('MainWindow', parent)
        self._game:Optional['Game'] = None
        super().__init__()
        Internal.scene = self

        self.scoreTable = None
        self.explainView = None
        self.clientDialog:Optional['ClientDialog']
        self.setupUi()
        assert Internal.Preferences
        Internal.Preferences.addWatch('showShadows', self.showShadowsChanged)
        Internal.Preferences.addWatch('animationSpeed', Speeds.refresh)

    @property
    def game(self) ->Optional['Game']:
        """a proxy"""
        return self._game

    @game.setter
    def game(self, value:Optional['Game']) ->None:
        """if it changes, update GUI"""
        changing = self._game != value
        game = self._game
        assert self.mainWindow
        self._game = value
        if changing:
            if value:
                self.mainWindow.updateGUI()
            else:
                assert game
                game.close()
                if self.scoreTable:
                    self.scoreTable.hide()
                if self.explainView:
                    self.explainView.hide()
                self.mainWindow.scene = None
        self.mainWindow.updateGUI()
        self.mainWindow.adjustMainView()

    @afterQueuedAnimations  # type:ignore[arg-type]
    def showShadowsChanged(self, deferredResult:'Deferred', # pylint: disable=unused-argument
        unusedOldValue:Any, unusedNewValue:Any) ->None:
        """if the wanted shadow direction changed, apply that change now"""
        for uiTile in self.graphicsTileItems():
            uiTile.setClippingFlags()
        self.applySettings()

    def handSelectorChanged(self, handBoard:'HandBoard') ->None:
        """update all relevant dialogs"""
        if self.game and not self.game.finished():
            assert handBoard.player
            handBoard.player.showInfo()
        # first decorate walls - that will compute player.handBoard for
        # explainView
        if self.explainView:
            self.explainView.refresh()

    def setupUi(self) ->None:
        """prepare scene"""
        assert Internal.Preferences
        assert isinstance(Internal.Preferences.windTilesetName, str)
        self.windTileset = Tileset(Internal.Preferences.windTilesetName)

    def showWall(self) ->None: # FIXME is unused
        """shows the wall according to the game rules (length may vary)"""
        if self.game:
            UIWall(self.game)   # sets self.game.wall

    def abort(self) ->Deferred:
        """abort current game"""
        # to be implemented by children
        return Deferred()

    def adjustSceneView(self) ->None:
        """adjust the view such that exactly the wanted things are displayed
        without having to scroll"""
        if self.game:
            if self.game.wall:
                with AnimationSpeed(99):
                    self.game.wall.decorate4()
                    for uiTile in self.game.wall.tiles:
                        _ = cast(UITile, uiTile)
                        if _.board:
                            _.board.placeTile(_)

    def applySettings(self) ->None:
        """apply preferences"""
        assert self.mainWindow
        assert Internal.Preferences
        self.mainWindow.actionAngle.setEnabled(
            bool(self.game) and bool(Internal.Preferences.showShadows))
        with AnimationSpeed():
            for item in self.nonTiles():
                if hasattr(item, 'tileset'):
                    item.tileset = Tileset.current()

    def prepareHand(self) ->None:
        """redecorate wall"""
        self.mainWindow.updateGUI()
        if self.game:
            assert self.game.wall
            with AnimationSpeed(Speeds.windDisc):
                self.game.wall.decorate4()

    def updateSceneGUI(self) ->None:
        """update some actions, all auxiliary windows and the statusbar"""
        game = self.game
        assert Internal.Preferences
        mainWindow = self.mainWindow
        if mainWindow:
            for action in [mainWindow.actionScoreGame, mainWindow.actionPlayGame]:
                action.setEnabled(not bool(game))
            mainWindow.actionAbortGame.setEnabled(bool(game))
            mainWindow.actionAngle.setEnabled(
                bool(game) and bool(Internal.Preferences.showShadows))
            for view in [self.explainView, self.scoreTable]:
                if view:
                    view.refresh()
            self.__showBalance()

    def newLightSource(self) ->Union[Literal['NE'], Literal['NW'], Literal['SW'], Literal['SE']]:
        """next value"""
        assert self.game
        assert self.game.wall
        oldIdx = LIGHTSOURCES.index(cast(UIWall, self.game.wall).lightSource)
        return cast(Union[Literal['NE'], Literal['NW'], Literal['SW'], Literal['SE']], LIGHTSOURCES[(oldIdx + 1) % 4])

    def changeAngle(self) ->None:
        """change the lightSource"""
        assert self.game
        assert self.game.wall
        cast(UIWall, self.game.wall).lightSource = self.newLightSource()
        self.focusRect.refresh()
        self.mainWindow.adjustMainView()

    def __showBalance(self) ->None:
        """show the player balances in the status bar"""
        sBar = self.mainWindow.statusBar()
        if self.game:
            for idx, player in enumerate(self.game.players):
                sbMessage = player.localName + ': ' + str(player.balance)
                if sBar.hasItem(idx):
                    sBar.changeItem(sbMessage, idx)
                else:
                    sBar.insertItem(sbMessage, idx, 1)
                    sBar.setItemAlignment(idx, Qt.AlignmentFlag.AlignLeft)
        else:
            for idx in range(5):
                if sBar.hasItem(idx):
                    sBar.removeItem(idx)

    def graphicsTileItems(self) ->Generator[UITile, None, None]:
        """return all UITile in the scene"""
        return (x for x in self.items() if isinstance(x, UITile))

    def nonTiles(self) ->Generator[QGraphicsItem, None, None]:
        """return all other items in the scene"""
        return (x for x in self.items() if not isinstance(x, UITile))

    def removeTiles(self) ->None:
        """remove all tiles from scene"""
        for item in self.graphicsTileItems():
            self.removeItem(item)
        for wind in Wind.all:
            if hasattr(wind, 'disc'):
                wind.disc.hide()
                delattr(wind, 'disc')
        self.focusRect.hide()
        UITile.clsUid = 0


class PlayingScene(GameScene):

    """scene with a playing game"""

    def __init__(self, parent:Optional['QWidget']=None) ->None:
        self._game = None
        self.__startingGame = True
        self._clientDialog:Optional['ClientDialog'] = None

        super().__init__(parent)

    @GameScene.game.setter  # type: ignore
    def game(self, value):
        game = self._game
        changing = value != game
        GameScene.game.fset(self, value)
        if changing:
            self.__startingGame = False
        self.mainWindow.actionChat.setEnabled(
            bool(value)
            and bool(value.client)
            and bool(value.client.connection)
            and not value.client.connection.url.isLocalGame)
        self.mainWindow.actionChat.setChecked(
            bool(value)
            and bool(value.client)
            and bool(value.client.table.chatWindow))

    @property
    def clientDialog(self) ->Optional['ClientDialog']:
        """wrapper: hide dialog when it is set to None"""
        return self._clientDialog

    @clientDialog.setter
    def clientDialog(self, value:Optional['ClientDialog']) ->None:
        """wrapper: hide dialog when it is set to None"""
        if self._clientDialog and isAlive(self._clientDialog) and not value:
            self._clientDialog.timer.stop()
            self._clientDialog.hide()
        self._clientDialog = value

    def setupUi(self) ->None:
        """create all other widgets
        we could make the scene view the central widget but I did
        not figure out how to correctly draw the background with
        QGraphicsView/QGraphicsScene.
        QGraphicsView.drawBackground always wants a pixmap
        for a huge rect like 4000x3000 where my screen only has
        1920x1200"""
        super().setupUi()
        self.setObjectName("PlayingField")

        self.discardBoard = DiscardBoard()
        self.addItem(self.discardBoard)

        self.adjustSceneView()

    def showWall(self) ->None:
        """shows the wall according to the game rules (length may vary)"""
        super().showWall()
        self.discardBoard.maximize()

    def abort(self) ->'Deferred':
        """abort current game"""
        def gotAnswer(gotResult:Any, autoPlaying:bool) ->Union[bool, 'Deferred']:
            """user answered"""
            result:Union[bool, 'Deferred'] = gotResult
            if result is True:
                self.game = None
            else:
                self.mainWindow.actionAutoPlay.setChecked(autoPlaying)
                result = False  # this implicitly handles the Cancelled Error
            return result
        if not self.game:
            return succeed(True)
        self.mainWindow.actionAutoPlay.setChecked(False)
        if self.game.finished():
            self.game = None
            return succeed(True)
        autoPlaying = self.mainWindow.actionAutoPlay.isChecked()
        return QuestionYesNo(i18n("Do you really want to abort this game?"), always=True).addBoth(
            gotAnswer, autoPlaying)

    def keyPressEvent(self, event:Optional['QKeyEvent']) ->None:
        """if we have a clientDialog, pass event to it"""
        if event:
            mod = event.modifiers()
            if mod in (Qt.KeyboardModifier.NoModifier, Qt.KeyboardModifier.ShiftModifier):
                if self.clientDialog:
                    self.clientDialog.keyPressEvent(event)
            super().keyPressEvent(event)

    def adjustSceneView(self) ->None:
        """adjust the view such that exactly the wanted things are displayed
        without having to scroll"""
        if self.game:
            with AnimationSpeed():
                self.discardBoard.maximize()
        super().adjustSceneView()

    @property
    def startingGame(self) ->bool:
        """are we trying to start a game?"""
        return self.__startingGame

    @startingGame.setter
    def startingGame(self, value:bool) ->None:
        """are we trying to start a game?"""
        if value != self.__startingGame:
            self.__startingGame = value
            self.mainWindow.updateGUI()

    def toggleDemoMode(self, checked:bool) ->None:
        """switch on / off for autoPlay"""
        if self.game:
            self.focusRect.refresh()  # show/hide it
            self.game.autoPlay = checked
            if checked and self.clientDialog:
                self.clientDialog.proposeAction()
                                                # an illegal action might have
                                                # focus
                self.clientDialog.selectButton()
                                               # select default, abort timeout

    def updateSceneGUI(self) ->None:
        """update some actions, all auxiliary windows and the statusbar"""
        if not isAlive(self):
            return
        super().updateSceneGUI()
        game = self.game
        mainWindow = self.mainWindow
        if not game:
            connections = [x.connection for x in HumanClient.humanClients if x.connection]
            title = ', '.join(f'{x.username}/{x.url}' for x in connections)
            if title:
                decorateWindow(mainWindow, title)
        else:
            decorateWindow(mainWindow, str(game.seed))
        for action in [mainWindow.actionScoreGame, mainWindow.actionPlayGame]:
            action.setEnabled(not bool(game))
        mainWindow.actionAbortGame.setEnabled(bool(game))
        self.discardBoard.setVisible(bool(game))
        mainWindow.actionAutoPlay.setEnabled(not self.startingGame)
        mainWindow.actionChat.setEnabled(bool(game) and bool(game.client)
                                         and bool(game.client.connection)
                                         and not game.client.connection.url.isLocalGame and not self.startingGame)
            # chatting on tables before game started works with chat button per
            # table
        mainWindow.actionChat.setChecked(
            mainWindow.actionChat.isEnabled(
            ) and bool(
                game.client.table.chatWindow))

    def changeAngle(self) ->None:
        """now that no animation is running, really change"""
        self.discardBoard.lightSource = self.newLightSource()
        super().changeAngle()


class ScoringScene(GameScene):

    """a scoring game"""

    def __init__(self, parent:Optional['QWidget']=None) ->None:
        self.scoringDialog:Optional[ScoringDialog] = None
        super().__init__(parent)
        self.selectorBoard.hasLogicalFocus = True

    @GameScene.game.setter  # type: ignore
    def game(self, value:'Game') ->None:
        game = self._game
        changing = value != game
        GameScene.game.fset(self, value)  # type:ignore[attr-defined]
        if changing:
            if value is not None:
                self.scoringDialog = ScoringDialog(scene=self)
            else:
                self.scoringDialog.hide()
                self.scoringDialog = None

    def handSelectorChanged(self, handBoard:'HandBoard') ->None:
        """update all relevant dialogs"""
        super().handSelectorChanged(handBoard)
        if self.scoringDialog:
            self.scoringDialog.slotInputChanged()

    def setupUi(self) ->None:
        """create all other widgets"""
        super().setupUi()
        self.setObjectName("ScoringScene")
        self.selectorBoard = SelectorBoard()
        self.addItem(self.selectorBoard)
        QMetaObject.connectSlotsByName(self)

    def abort(self) ->'Deferred':
        """abort current game"""
        def answered(result:'Deferred') ->'Deferred':
            """got answer"""
            if result:
                self.game = None
            return result
        if Debug.quit:
            logDebug('ScoringScene.abort invoked')
        if not self.game:
            return succeed(True)
        if self.game.finished():
            self.game = None
            return succeed(True)
        return QuestionYesNo(i18n("Do you really want to abort this game?"), always=True).addCallback(
            answered).addErrback(logFailure)

    def __moveTile(self, uiTile:UITile, wind:Wind, toConcealed:bool) ->None:
        """the user pressed a wind letter or X for center, wanting to move a uiTile there"""
        # this tells the receiving board that this is keyboard, not mouse navigation>
        # needed for useful placement of the popup menu
        if wind is NoWind:
            receiver = self.selectorBoard
        else:
            receiver = self.game.players[wind].handBoard
        currentBoard = uiTile.board
        assert currentBoard
        movingMeld = currentBoard.uiMeldWithTile(uiTile)
        if receiver != currentBoard or toConcealed != movingMeld.meld.isConcealed:
            movingLastMeld = movingMeld.meld == self.computeLastMeld()
            assert self.scoringDialog
            if movingLastMeld:
                self.scoringDialog.clearLastTileCombo()
            receiver.dropMeld(uiTile, toConcealed)
            if movingLastMeld and receiver == currentBoard:
                self.scoringDialog.fillLastTileCombo()

    def __navigateScoringGame(self, event:'QKeyEvent') ->bool:
        """keyboard navigation in a scoring game"""
        key = event.key()
        uiTile = cast(UITile, self.focusItem())
        if uiTile is None:
            logError(f'received key {key}/{event.text()} but focusItem() is None, '
                     f'scene.activePanel: {self.activePanel()}')
            return False
        wind = Wind.normalized(key)
        if wind is not None:
            shift = bool(event.modifiers() & Qt.KeyboardModifier.ShiftModifier)
            self.__moveTile(uiTile, wind, shift)
            return True
        if key == Qt.Key.Key_Tab and self.game:
            tabItems = [self.selectorBoard]
            tabItems.extend(p.handBoard for p in self.game.players if not p.handBoard.empty)
            tabItems.append(tabItems[0])
            currentBoard = uiTile.board  # type: ignore[attr-defined]
            currIdx = 0
            while tabItems[currIdx] != currentBoard and currIdx < len(tabItems) - 2:
                currIdx += 1
            tabItems[currIdx + 1].hasLogicalFocus = True
            return True
        return False

    def keyPressEvent(self, event:Optional['QKeyEvent']) ->None:
        """navigate in the selectorboard"""
        if event:
            mod = event.modifiers()
            if mod in (Qt.KeyboardModifier.NoModifier, Qt.KeyboardModifier.ShiftModifier):
                if self.game:
                    if self.__navigateScoringGame(event):
                        return
            super().keyPressEvent(event)

    def adjustSceneView(self) ->None:
        """adjust the view such that exactly the wanted things are displayed
        without having to scroll"""
        if self.game:
            with AnimationSpeed():
                self.selectorBoard.maximize()
        super().adjustSceneView()

    def prepareHand(self) ->None:
        """redecorate wall"""
        super().prepareHand()
        if self.scoringDialog:
            self.scoringDialog.clearLastTileCombo()

    def updateSceneGUI(self) ->None:
        """update some actions, all auxiliary windows and the statusbar"""
        if not isAlive(self):
            return
        super().updateSceneGUI()
        game = self.game
        mainWindow = self.mainWindow
        for action in [mainWindow.actionScoreGame, mainWindow.actionPlayGame]:
            action.setEnabled(not bool(game))
        mainWindow.actionAbortGame.setEnabled(bool(game))
        self.selectorBoard.setVisible(bool(game))
        self.selectorBoard.setEnabled(bool(game))

    def changeAngle(self) ->None:
        """now that no animation is running, really change"""
        self.selectorBoard.lightSource = self.newLightSource()
        super().changeAngle()

    def computeLastTile(self) ->'Tile':
        """compile hand info into a string as needed by the scoring engine"""
        if self.scoringDialog:
            # is None while ScoringGame is created
            return self.scoringDialog.computeLastTile()
        return Tile.none

    def computeLastMeld(self) ->Meld:
        """compile hand info into a string as needed by the scoring engine"""
        if self.scoringDialog:
            # is None while ScoringGame is created
            cbLastMeld = self.scoringDialog.cbLastMeld
            idx = cbLastMeld.currentIndex()
            if idx >= 0:
                return Meld(cbLastMeld.itemData(idx))
        return Meld()



# -*- coding: utf-8 -*-

"""
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

# pylint: disable=wrong-import-order

from kdestub import *  # pylint: disable=wildcard-import,unused-wildcard-import



# -*- coding: utf-8 -*-

"""
Copyright (C) 2010-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

import os
import sys
import tarfile
import subprocess
import datetime
from io import BytesIO
from hashlib import md5
from typing import List, Any, Optional

from common import Debug, Internal, ReprMixin, cacheDir
from util import which, removeIfExists, uniqueList, elapsedSince
from log import logWarning, i18n, logDebug, logException

from qt import QStandardPaths

from tile import Tile

        # Phonon does not work with short files - it plays them
        # simultaneously or only parts of them. Mar 2010, KDE 4.4. True for mp3
        # and for wav. Also, mpg123 often plays distorted sounds. Kubuntu 9.10.
        # So we use ogg123 and ogg sound files.
        # self.audio = Phonon.MediaObject(self)
        # self.audioOutput = Phonon.AudioOutput(Phonon.GameCategory, self)
        # Phonon.createPath(self.audio, self.audioOutput)
        # self.audio.enqueue(Phonon.MediaSource(wavName))
        # self.audio.play()

if sys.platform == 'win32':
    import winsound  # pylint: disable=import-error


class SoundPopen(subprocess.Popen):

    """with additional attributes"""

    def __init__(self, what:str, args:Any) ->None:
        super().__init__(args)
        self.name = what
        self.startTime = datetime.datetime.now()


class Sound:

    """the sound interface. Use class variables and class methods,
    thusly ensuring no two instances try to speak"""
    __oggBinary = None
    __bonusOgg = None
    playProcesses : List[SoundPopen] = []
    lastCleaned = None

    @staticmethod
    def findOggBinary() ->str:
        """set __oggBinary to exe name or an empty string"""
        if Sound.__oggBinary is None:
            if sys.platform == 'win32':
                Sound.__oggBinary = os.path.join('share', 'kajongg', 'voices', 'oggdec.exe')
                msg = ''  # we bundle oggdec.exe with the kajongg installer, it must be there
            else:
                oggBinary = 'ogg123'
                msg = i18n(
                    'No voices will be heard because the program %1 is missing',
                    oggBinary)
                if which(oggBinary):
                    Sound.__oggBinary = oggBinary
                else:
                    Sound.__oggBinary = ''
                    assert Internal.Preferences
                    Internal.Preferences.useSounds = False
                    # checks again at next reenable
                    if msg:
                        logWarning(msg)
            if Debug.sound:
                logDebug('ogg123 found:' + Sound.__oggBinary)
        return Sound.__oggBinary

    @staticmethod
    def cleanProcesses() ->None:
        """terminate ogg123 children"""
        now = datetime.datetime.now()
        if Sound.lastCleaned and (now - Sound.lastCleaned).seconds < 2:
            return
        Sound.lastCleaned = now
        remaining = []
        for process in Sound.playProcesses:
            if process.poll() is not None:
                # ogg123 self-finished
                continue
            diff = now - process.startTime
            if diff.seconds > 10:
                try:
                    process.kill()
                except OSError:
                    pass
                try:
                    os.waitpid(process.pid, 0)
                except OSError:
                    pass
                if Debug.sound:
                    assert Internal.scene
                    game = Internal.scene.game
                    assert game
                    game.debug(f'10 seconds passed. Killing {process.name}')
            else:
                remaining.append(process)
        Sound.playProcesses = remaining

    @staticmethod
    def speak(what:str) ->None:
        """this is what the user of this module will call."""
        # pylint: disable=too-many-branches
        if not Internal.scene:
            return
        assert Internal.Preferences
        if not Internal.Preferences.useSounds:
            return
        game = Internal.scene.game
        if not game:
            return
        reactor = Internal.reactor
        assert reactor
        if game and not game.autoPlay and len(Sound.playProcesses) > 0:
            # in normal play, wait a moment between two speaks. Otherwise
            # sometimes too many simultaneous speaks make them ununderstandable
            lastSpeakStart = max(x.startTime for x in Sound.playProcesses)
            if elapsedSince(lastSpeakStart) < 0.3:
                assert reactor
                reactor.callLater(1, Sound.speak, what)
                return
        if os.path.exists(what):
            oggBinary = Sound.findOggBinary()
            if oggBinary:
                if sys.platform == 'win32':
                    # convert to .wav, store .wav in cacheDir
                    name, ext = os.path.splitext(what)
                    assert ext == '.ogg', f'what: {what} name: {name} ext: {ext}'
                    if 'bell' in name:
                        nameParts = ['bell']
                    else:
                        nameParts = os.path.normpath(name).split(os.sep)
                        nameParts = nameParts[nameParts.index('voices') + 1:]
                    wavName = os.path.normpath(
                        f"{cacheDir()}/{'_'.join(nameParts)}.wav")
                    if not os.path.exists(wavName):
                        args = [oggBinary, '-a', '-w', wavName, os.path.normpath(what)]
                        startupinfo = subprocess.STARTUPINFO()
                        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
                        subprocess.call(args, startupinfo=startupinfo)
                        if Debug.sound:
                            logDebug(f'converted {what} to wav {wavName}')
                    try:
                        winsound.PlaySound(
                            wavName,
                            winsound.SND_FILENAME | winsound.SND_NODEFAULT)
                    except RuntimeError:
                        pass
                else:
                    args = [oggBinary, '-q', what]
                    if Debug.sound:
                        game.debug(' '.join(args))
                    process = SoundPopen(what, args)
                    Sound.playProcesses.append(process)
                    reactor.callLater(3, Sound.cleanProcesses)
                    reactor.callLater(6, Sound.cleanProcesses)

    @staticmethod
    def bonus() ->None:
        """ring some sort of bell, if we find such a file"""
        if Sound.__bonusOgg is None:
            Sound.__bonusOgg = ''
            for oggName in (
                    '/usr/share/sounds/KDE-Im-Message-In.ogg',
                    'share/sounds/bell.ogg'):
                if os.path.exists(oggName):
                    Sound.__bonusOgg = oggName
                    if Debug.sound:
                        logDebug(f'Bonus sound found:{oggName}')
                    break
            if Debug.sound and not Sound.__bonusOgg:
                logDebug('No bonus sound found')
        if Sound.__bonusOgg:
            Sound.speak(Sound.__bonusOgg)

class Voice(ReprMixin):

    """this administers voice sounds.

    When transporting voices between players, a compressed tarfile
    is generated at source and transferred to destination. At
    destination, no tarfile is written, only the content. It makes
    only sense to cache the voice in a tarfile at source."""

    __availableVoices : List['Voice'] = []
    md5sumLength = 32 # magical constant

    def __init__(self, directory:str, content:Optional[bytes]=None) ->None:
        """give this name a voice"""
        self.__md5sum:Optional[str] = None
        if not os.path.split(directory)[0]:
            if Debug.sound:
                logDebug(f'place voice {directory} in {cacheDir()}')
            directory = os.path.join(cacheDir(), directory)
        self.directory = directory
        self.__setArchiveContent(content)

    def __str__(self) ->str:
        return self.directory

    def language(self) ->str:
        """the language code of this voice. Locally defined voices
        have no language code and return 'local'.
        remote voices received from other clients return 'remote',
        they always get predecence."""
        if len(self.directory) == self.md5sumLength:
            if os.path.split(self.directory)[1] == self.md5sum:
                return 'remote'
        if 'HOME' in os.environ:
            home = os.environ['HOME']
        elif 'HOMEPATH' in os.environ:
            home = os.environ['HOMEPATH']
        else:
            logWarning('have neither HOME nor HOMEPATH')
            home = ''
        if home:
            if self.directory.startswith(home):
                return 'local'
        result = os.path.split(self.directory)[0]
        result = os.path.split(result)[1]
        if result == 'voices':
            result = 'en_US'
        return result

    @staticmethod
    def availableVoices() ->List['Voice']:
        """a list of all voice directories"""
        if not Voice.__availableVoices:
            result:List['Voice'] = []
            directories = QStandardPaths.locateAll(
                QStandardPaths.StandardLocation.AppDataLocation, 'voices', QStandardPaths.LocateOption.LocateDirectory)
            directories.insert(0, os.path.join('share', 'kajongg', 'voices'))
            for parentDirectory in directories:
                for (dirpath, _, _) in os.walk(parentDirectory, followlinks=True):
                    if os.path.exists(os.path.join(dirpath, 's1.ogg')):
                        result.append(Voice(dirpath))
            group = Internal.kajonggrc.group('Locale')
            _ = uniqueList(
                ':'.join(['local', str(group.readEntry('Language')), 'en_US']).split(':'))
            prefLanguages = dict((x[1], x[0]) for x in enumerate(_))
            result = sorted(
                result, key=lambda x: prefLanguages.get(x.language(), 9999))
            if Debug.sound:
                logDebug(f'found voices:{[str(x) for x in result]}')
            Voice.__availableVoices = result
        return Voice.__availableVoices

    @staticmethod
    def locate(name:str) ->Optional['Voice']:
        """return Voice or None if no foreign or local voice matches.
        In other words never return a predefined voice"""
        for voice in Voice.availableVoices():
            dirname = os.path.split(voice.directory)[-1]
            if name == voice.md5sum:
                if Debug.sound:
                    logDebug(
                        f'locate found {name} by md5sum in {voice.directory}')
                return voice
            if name == dirname and voice.language() == 'local':
                if Debug.sound:
                    logDebug(
                        f'locate found {name} by name in {voice.directory}')
                return voice
        if Debug.sound:
            logDebug(f'Personal sound for {name} not found')
        return None

    def buildSubvoice(self, oggName:str, side:str) ->str:
        """side is 'left' or 'right'."""
        angleDirectory = os.path.join(
            cacheDir(),
            'angleVoices',
            self.md5sum,
            side)
        stdName = os.path.join(self.directory, oggName)
        angleName = os.path.join(angleDirectory, oggName)
        if os.path.exists(stdName) and not os.path.exists(angleName):
            sox = which('sox')
            if not sox:
                return stdName
            if not os.path.exists(angleDirectory):
                os.makedirs(angleDirectory)
            args = [sox, stdName, angleName, 'remix']
            if side == 'left':
                args.extend(['1,2', '0'])
            elif side == 'right':
                args.extend(['0', '1,2'])
            callResult = subprocess.call(args)
            if callResult:
                if Debug.sound:
                    logDebug(
                        f'failed to build subvoice {angleName}: return code={callResult}')
                return stdName
            if Debug.sound:
                logDebug(f'built subvoice {angleName}')
        return angleName

    def localTextName(self, text:str, angle:float) ->str:
        """build the name of the wanted sound file"""
        oggName = text.lower().replace(' ', '') + '.ogg'
        if angle == 90:
            return self.buildSubvoice(oggName, 'left')
        if angle == 270:
            return self.buildSubvoice(oggName, 'right')
        return os.path.join(self.directory, oggName)

    def speak(self, text:str, angle:float) ->None:
        """text must be a sound filename without extension"""
        if isinstance(text, Tile):
            text = str(text.exposed)
        fileName = self.localTextName(text, angle)
        if not os.path.exists(fileName):
            if Debug.sound:
                logDebug(f'Voice.speak: fileName {fileName} not found')
        Sound.speak(fileName)

    def oggFiles(self) ->List[str]:
        """a list of all found ogg files"""
        if os.path.exists(self.directory):
            return sorted(x for x in os.listdir(self.directory) if x.endswith('.ogg'))
        return []

    def __computeMd5sum(self) ->None:
        """update md5sum file. If it changed, return True.
        If unchanged or no ogg files exist, remove archive and md5sum and return False.
        If ogg files exist but no archive, return True."""
        if self.__md5sum:
            # we already checked
            return
        md5FileName = os.path.join(self.directory, 'md5sum')
        archiveExists = os.path.exists(self.archiveName())
        ogg = self.oggFiles()
        if not ogg:
            removeIfExists(self.archiveName())
            removeIfExists(md5FileName)
            self.__md5sum = None
            logDebug(f'no ogg files in {self}')
            return
        md5sum = md5()
        for oggFile in ogg:
            with open(os.path.join(self.directory, oggFile), 'rb') as _:
                md5sum.update(_.read())
        # the md5 stamp goes into the old archive directory 'username'
        self.__md5sum = md5sum.hexdigest()
        existingMd5sum = self.savedmd5Sum()
        md5Name = self.md5FileName()
        if self.__md5sum != existingMd5sum:
            if Debug.sound:
                if not os.path.exists(md5Name):
                    logDebug(f'creating new {md5Name}')
                else:
                    logDebug(
                        f'md5sum {existingMd5sum} changed, rewriting {md5Name} with {self.__md5sum}')
            try:
                with open(md5Name, 'w', encoding='ascii') as _:
                    _.write(f'{self.__md5sum}\n')
            except OSError as exception:
                logException(
                    '\n'.join([i18n('cannot write <filename>%1</filename>: %2',
                                    md5Name,
                                    str(exception)),
                               i18n('The voice files have changed, their checksum has changed.'),
                               i18n('Please reinstall kajongg or do, with sufficient permissions:'),
                               f'cd {self.directory} ; cat *.ogg | md5sum > md5sum']))
        if archiveExists:
            archiveIsOlder = os.path.getmtime(
                md5Name) > os.path.getmtime(self.archiveName())
            if self.__md5sum != existingMd5sum or archiveIsOlder:
                os.remove(self.archiveName())

    def __buildArchive(self) ->None:
        """write the archive file and set self.__md5sum"""
        self.__computeMd5sum()
        if not os.path.exists(self.archiveName()):
            with tarfile.open(self.archiveName(), mode='w:bz2') as tarFile:
                for oggFile in self.oggFiles():
                    tarFile.add(
                        os.path.join(
                            self.directory,
                            oggFile),
                        arcname=oggFile)

    def archiveName(self) ->str:
        """ the full path of the archive file"""
        return os.path.join(self.directory, 'content.tbz')

    def md5FileName(self) ->str:
        """the name of the md5sum file"""
        return os.path.join(self.directory, 'md5sum')

    def savedmd5Sum(self) ->Optional[str]:
        """return the current value of the md5sum file"""
        if os.path.exists(self.md5FileName()):
            try:
                with open(self.md5FileName(), 'r', encoding='ascii') as _:
                    line = _.readlines()[0].replace(' -', '').strip()
                if len(line) == self.md5sumLength:
                    return line
                logWarning(f'{self.md5FileName()} has wrong content: {line}')
            except OSError as exc:
                logWarning(f'{self.md5FileName()} has wrong content: {exc}')
        return None

    @property
    def md5sum(self) ->str:
        """the current checksum over all ogg files"""
        self.__computeMd5sum()
        assert self.__md5sum
        return self.__md5sum

    def __setArchiveContent(self, content:Optional[bytes]) ->None:
        """fill the Voice with ogg files"""
        if not content:
            return
        if not os.path.exists(self.directory):
            os.makedirs(self.directory)
        filelike = BytesIO(content)
        with tarfile.open(mode='r|bz2', fileobj=filelike) as tarFile:
            tarFile.extractall(path=self.directory)
            if Debug.sound:
                logDebug(f'extracted archive into {self.directory}')
        filelike.close()

    @property
    def archiveContent(self) ->Optional[bytes]:
        """the content of the tarfile"""
        self.__buildArchive()
        if os.path.exists(self.archiveName()):
            return open(self.archiveName(), 'rb').read()
        return None

    @archiveContent.setter
    def archiveContent(self, content:Optional[bytes]) ->None:
        """new archive content"""
        self.__setArchiveContent(content)



# -*- coding: utf-8 -*-

"""
Copyright (C) 2013-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

from typing import List, Any, Optional, Tuple, TYPE_CHECKING, cast

from qt import QColor

from mi18n import i18nc
from message import Message
from common import Internal, isAlive
from player import Player, PlayingPlayer
from game import PlayingGame
from tile import Tile, TileList
from handboard import PlayingHandBoard
from animation import AnimationSpeed
from uiwall import UIWall, SideText
from wind import Wind

if TYPE_CHECKING:
    from tile import Tiles, Meld, MeldList
    from uitile import UITile
    from uiwall import UIWallSide
    from hand import Hand
    from rule import Ruleset
    from client import Client
    from scene import PlayingScene

class VisiblePlayer(Player):

    """Mixin for VisiblePlayingPlayer and ScoringPlayer"""

    def __init__(self) ->None:
        # pylint: disable=super-init-not-called
        assert self.game
        assert self.game.wall
        self.__front = cast(UIWall, self.game.wall)[self.idx]
        self.sideText = SideText()
        self.sideText.board = self.__front

    def hide(self) ->None:
        """clear visible data and hide"""
        self.clearHand()
        if isAlive(self.handBoard):
            assert self.handBoard
            self.handBoard.hide()

    @property
    def idx(self) ->int:
        """our index in the player list"""
        assert self.game
        if self not in self.game.players:
            # we will be added next
            return len(self.game.players)
        return self.game.players.index(self)

    @property
    def front(self) ->'UIWallSide':
        """front"""
        return self.__front

    @front.setter
    def front(self, value:'UIWallSide') ->None:
        """also assign handBoard to front"""
        self.__front = value
        if self.handBoard:
            self.handBoard.setParentItem(value)

    def syncHandBoard(self, adding:Optional[List['UITile']]=None) ->None:
        """update display of handBoard. Set Focus to tileName."""
        assert self.handBoard
        self.handBoard.sync(adding)

    def showInfo(self) ->None:
        """show player info on the wall"""
        side = self.front
        self.sideText.text = f'{self.localName} - {self.explainHand().total()}'  # type:ignore[attr-defined]
        self.colorizeName()
        myWind = Wind.all4[self.wind]
        if hasattr(myWind, 'disc'):
            # sometimes this is not the case. TODO: when/why?
            _ = myWind.disc
            assert _
            side.disc = _
        assert self.game
        side.disc.is_prevailing = self.wind == self.game.point.prevailing
        side.disc.board = self.front


class VisiblePlayingPlayer(VisiblePlayer, PlayingPlayer):

    """this player instance has a visual representation"""

    def __init__(self, game:'VisiblePlayingGame', name:str) ->None:
        assert game
        self.handBoard = None  # because Player.init calls clearHand()
        PlayingPlayer.__init__(self, game, name)
        VisiblePlayer.__init__(self)
        self.handBoard = PlayingHandBoard(self)
        self.voice = None

    def clearHand(self) ->None:
        """clears attributes related to current hand"""
        super().clearHand()
        if self.game and self.game.wall:
            # is None while __del__
            self.front = cast(UIWall, self.game.wall)[self.idx]
        if self.handBoard:
            self.handBoard.setEnabled(
                self.game is not None and self.game.belongsToHumanPlayer(
                ) and self == self.game.myself)

    def explainHand(self) ->'Hand':
        """return the hand to be explained. Same as current unless we need to discard.
        In that case, make an educated guess about the discard.
        For player==game.myself, use the focused tile."""
        hand = self.hand
        assert self.handBoard
        if hand and hand.tiles and self._concealedTiles:
            if hand.lenOffset == 1 and not hand.won:
                if any(not x.isKnown for x in self._concealedTiles):
                    hand -= Tile.unknown
                elif self.handBoard.focusTile:
                    hand -= self.handBoard.focusTile.tile
        return hand

    def colorizeName(self) ->None:
        """set the color to be used for showing the player name on the wall"""
        if not isAlive(self.sideText):
            return
        assert self.game
        assert Internal.Preferences
        if self == self.game.activePlayer and self.game.client:
            color = 'blue'
        elif Internal.Preferences.tilesetName == 'jade':
            color = 'white'
        else:
            color = 'black'
        self.sideText.color = QColor(color)

    def getsFocus(self, unusedResults:Any=None) ->None:
        """give this player focus on his handBoard"""
        assert self.handBoard
        self.handBoard.setEnabled(True)
        self.handBoard.hasLogicalFocus = True

    def popupMsg(self, msg:Message) ->None:
        """shows a yellow message from player"""
        if msg != Message.NoClaim:
            self.speak(msg.name.lower())
            yellow = self.front.message
            yellow.setText(f"{yellow.msg}  {i18nc('kajongg', msg.name)}")
            yellow.setVisible(True)

    def hidePopup(self) ->None:
        """hide the yellow message from player"""
        if isAlive(self.front.message):
            self.front.message.msg = ''
            self.front.message.setVisible(False)

    def speak(self, txt:str) ->None:
        """speak if we have a voice"""
        if self.voice:
            self.voice.speak(txt, self.front.rotation())

    def getsRobbed(self, tile:Tile) ->None:
        """used for robbing the kong from this player"""
        super().getsRobbed(tile)
        assert self.handBoard
        lastDiscard = self.handBoard[tile.exposed][-1]  # type:ignore[index]
        lastDiscard.change_name(lastDiscard.concealed)
        cast('PlayingScene', Internal.scene).discardBoard.discardTile(lastDiscard)
        assert lastDiscard.isConcealed
        self.syncHandBoard()

    def addConcealedTiles(self, tiles:'Tiles', animated:bool=True) ->None:
        """add to my tiles and sync the hand board"""
        assert Internal.Preferences
        _ = cast(List['UITile'], tiles)
        with AnimationSpeed(speed=int(Internal.Preferences.animationSpeed) if animated else 99):
            super().addConcealedTiles(TileList(x.tile for x in _))
            self.syncHandBoard(_)

    def declaredMahJongg(self, concealed:'MeldList', withDiscard:Optional[Tile], lastTile:Tile, lastMeld:'Meld') ->None:
        """player declared mah jongg. Determine last meld, show
        concealed tiles grouped to melds"""
        super().declaredMahJongg(concealed, withDiscard, lastTile, lastMeld)
        if withDiscard:
            # withDiscard is a Tile, we need the UITile
            discardedTile = cast('PlayingScene', Internal.scene).discardBoard.claimDiscard()
            if discardedTile.tile is not withDiscard:
                assert self.game
                self.game.debug(
                    f'{discardedTile.tile} is not {withDiscard}')
                assert False
            self.syncHandBoard([discardedTile])
        else:
            # show concealed tiles
            self.syncHandBoard()

    def removeConcealedTile(self, tile:Tile) ->None:
        """remove from my melds or tiles"""
        super().removeConcealedTile(tile)
        self.syncHandBoard()

    def makeTileKnown(self, tile:Tile) ->None:
        """give an unknown tileItem a name"""
        super().makeTileKnown(tile)
        assert tile.isKnown
        assert self.handBoard
        matchingTiles = sorted(
            self.handBoard.tilesByElement(Tile.unknown),
            key=lambda x: x.xoffset)
        matchingTiles[-1].change_name(tile)

    def exposeMeld(self, meldTiles:'Tiles', calledTile:Optional[Tile]=None) ->'Meld':
        result = PlayingPlayer.exposeMeld(
            self,
            meldTiles,
            cast('UITile', calledTile).tile if calledTile else None)  # type:ignore[arg-type]
        adding = [cast('UITile', calledTile)] if calledTile else None
        self.syncHandBoard(adding=adding)
        return result


class VisiblePlayingGame(PlayingGame):

    """for the client"""
    playerClass = VisiblePlayingPlayer
    wallClass = UIWall

    def __init__(self, names:List[Tuple['Wind', str]], ruleset:'Ruleset', gameid:Optional[int]=None,
                 wantedGame:Optional[str]=None, client:Optional['Client']=None,
                 playOpen:bool=False, autoPlay:bool=False) ->None:
        PlayingGame.__init__(
            self, names, ruleset, gameid, wantedGame=wantedGame,
            client=client, playOpen=playOpen, autoPlay=autoPlay)
#        Internal.mainWindow.adjustMainView()
#        Internal.mainWindow.updateGUI()
        assert self.wall
        self.wall.decorate4()

    def close(self) ->None:
        """close the game"""
        scene = cast('PlayingScene', Internal.scene)
        assert scene
        scene.discardBoard.hide()
        if isAlive(scene):
            scene.removeTiles()
        scene.clientDialog = None
        for player in self.players:
            player.hide()
        if self.wall:
            self.wall.hide()
        if isAlive(scene.mainWindow):
            scene.mainWindow.actionAutoPlay.setChecked(False)
        scene.startingGame = False
        scene.game = None
        SideText.removeAll()
        scene.mainWindow.updateGUI()
        return super().close()
